#------------------------------------------------------------------------------
# Name:         test_nansat.py
# Purpose:      Test the Nansat class
#
# Author:       Morten Wergeland Hansen, Asuka Yamakawa
# Modified: Morten Wergeland Hansen
#
# Created:      18.06.2014
# Last modified:16.04.2015 10:48
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
#------------------------------------------------------------------------------
from __future__ import unicode_literals, absolute_import
import os
import unittest
import datetime
import warnings

try:
    if 'DISPLAY' not in os.environ:
        import matplotlib; matplotlib.use('Agg')
    from matplotlib.colors import hex2color
except ImportError:
    MATPLOTLIB_IS_INSTALLED = False
else:
    MATPLOTLIB_IS_INSTALLED = True

from nansat.warnings import NansatFutureWarning
from nansat.tools import get_random_color, parse_time
from nansat.tools import (OptionError,
                            ProjectionError,
                            GDALError,
                            NansatReadError,
                            GeolocationError,
                            WrongMapperError)

class ToolsTest(unittest.TestCase):
    @unittest.skipUnless(MATPLOTLIB_IS_INSTALLED, 'Matplotlib is required')
    def test_get_random_color(self):
        ''' Should return HEX code of random color '''
        c0 = get_random_color()
        c1 = get_random_color(c0)
        c2 = get_random_color(c1, 300)

        self.assertEqual(type(hex2color(c0)), tuple)
        self.assertEqual(type(hex2color(c1)), tuple)
        self.assertEqual(type(hex2color(c2)), tuple)

    def test_parse_time(self):
        dt = parse_time('2016-01-19')

        self.assertEqual(type(dt), datetime.datetime)

    def test_parse_time_incorrect(self):
        dt = parse_time('2016-01-19Z')

        self.assertEqual(type(dt), datetime.datetime)

    def test_OptionError_warning(self):
        with warnings.catch_warnings(record=True) as recorded_warnings:
            with self.assertRaises(OptionError):
                raise OptionError
            self.assertEqual(recorded_warnings[0].category, NansatFutureWarning)

    def test_ProjectionError_warning(self):
        with warnings.catch_warnings(record=True) as recorded_warnings:
            with self.assertRaises(ProjectionError):
                raise ProjectionError
            self.assertEqual(recorded_warnings[0].category, NansatFutureWarning)

    def test_GDALError_warning(self):
        with warnings.catch_warnings(record=True) as recorded_warnings:
            with self.assertRaises(GDALError):
                raise GDALError
            self.assertEqual(recorded_warnings[0].category, NansatFutureWarning)

    def test_NansatReadError_warning(self):
        with warnings.catch_warnings(record=True) as recorded_warnings:
            with self.assertRaises(NansatReadError):
                raise NansatReadError
            self.assertEqual(recorded_warnings[0].category, NansatFutureWarning)

    def test_GeolocationError_warning(self):
        with warnings.catch_warnings(record=True) as recorded_warnings:
            with self.assertRaises(GeolocationError):
                raise GeolocationError
            self.assertEqual(recorded_warnings[0].category, NansatFutureWarning)

    def test_WrongMapperError_warning(self):
        with warnings.catch_warnings(record=True) as recorded_warnings:
            with self.assertRaises(WrongMapperError):
                raise WrongMapperError
            self.assertEqual(recorded_warnings[0].category, NansatFutureWarning)




import unittest

from nansat.nansat import Nansat
from nansat.domain import Domain
from nansat.nsr import NSR

doppler_installed = True
try:
    from sardoppler.sardoppler import Doppler
except Exception as e:
    print e.message
    doppler_installed = False

class TestOpenIssues(unittest.TestCase):

    def test_issue_189(self):
        fn = '/mnt/10.11.12.232/sat_downloads_asar/level-0/2010-01/descending/VV/gsar_rvl/RVL_ASA_WS_20100110211812087.gsar'
        if doppler_installed:
            n = Doppler(fn)
            xlon, xlat = n.get_corners()
            d = Domain(NSR(3857),
                    '-lle %f %f %f %f -tr 1000 1000' % (
                        xlon.min(), xlat.min(), xlon.max(), xlat.max()))
            n.reproject(d, eResampleAlg=1, tps=True)
            inci = n['incidence_angle']




# Name:    pointbrowser.py
# Purpose: contains PointBrowser class
# Authors:      Asuka Yamakawa, Anton Korosov, Knut-Frode Dagestad,
#               Morten W. Hansen, Alexander Myasoyedov,
#               Dmitry Petrenko, Evgeny Morozov,
#               Aleksander Vines
# Created:      29.06.2011
# Copyright:    (c) NERSC 2011 - 2015
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
import os
import numpy as np

try:
    import matplotlib
    if 'DISPLAY' not in os.environ:
        matplotlib.use('Agg')
    import matplotlib.pyplot as plt
except ImportError:
    MATPLOTLIB_IS_INSTALLED = False
else:
    MATPLOTLIB_IS_INSTALLED = True


class PointBrowser():
    """
    Click on raster images shown by plt.imshow and get the X-Y coordinates.

    Parameters
    ----------
    data : ndarray
        image to imshow
    transect : bool
        if True, get transects / points
        if False, get only points
    force_interactive : bool
        force PointBrowser to interactive mode? (True for regular use, False for tests)
    **kwargs : dict
        optional parameters for imshow

    Note
    ----
    self.fig        : pyplot Figure
    self.data       : ndarray with data
    self.ax         : axes
    self.points     : plot with points
    self.line       : plot with points
    self.coordinates: container for recorded coordinates

    """
    # instance attributes
    fig = None
    data = None
    fmt = None
    text_ax = None
    ax = None
    points = None
    lines = None
    coordinates = None

    def __init__(self, data, fmt='x-k', force_interactive=True, **kwargs):
        """Open figure with imshow and colorbar"""
        if not MATPLOTLIB_IS_INSTALLED:
            raise ImportError(' Matplotlib is not installed ')
        if force_interactive and not matplotlib.is_interactive():
            raise SystemError('''
        Python is started with -pylab option, transect will not work.
        Please restart python without -pylab.''')

        self.fig = plt.figure()
        self.data = data
        self.fmt = fmt
        self.text_ax = plt.axes([0.0, 0.85, 1.0, 0.15])
        self.ax = plt.axes([0.0, 0.0, 1.0, 0.85])
        img = self.ax.imshow(self.data, extent=(0, self.data.shape[1],
                                                0, self.data.shape[0]),
                             origin='lower', **kwargs)

        self.fig.colorbar(img)
        self.points = []
        self.lines = [self.ax.plot([], [], self.fmt)[0]]
        self.coordinates = [[]]

    def onclick(self, event):
        """ Process mouse onclick event
        Append coordinates of the click to self.coordinates, add point and 2D line to self.points
        If click is outside, nothing is done
        If click with 'z' pressed, nothing is done
        If click with 'anykey', new line is started

        Parameters
        ----------
        event : matplotlib.mouse_event

        """
        # ignore click outside image
        if event.xdata is None or event.ydata is None:
            return

        # ignore clicked point if "z" key is held down
        if str(event.key) == 'alt+z' or str(event.key) == 'z':
            return

        if event.key is not None:
            # - holding down any other key (NOT cmd (mac),shift,alt,ctrl)
            #   means a new line is started at the clicked point
            self.coordinates.append([])
            self.lines.append(self.ax.plot([], [], self.fmt)[0])

        self.coordinates[-1].append((event.xdata, event.ydata))
        self.points.append(self.ax.plot(event.xdata, event.ydata, self.fmt))
        self.lines[-1].set_data(np.array(self.coordinates[-1]).T)
        self.ax.figure.canvas.draw()

    def _convert_coordinates(self):
        """ Converts the coordinates array to points array for return in
        get_points, so that the internal structure can be tested

        Returns
        -------
            list of arrays
            The format of the returned array:
            [array([[x1,...,xn],[y1,...,yn]]),array([[xn+1,...],[yn+1,...]]),...]
            Each 'array' element is a numpy.ndarray and represents one transect,
            where x1,y1 is the first point in the first transect,
            and xn,yn the last point in the first transect.
            The inner x/y-arrays are also numpy.ndarrays
        """
        return [np.array(p).T for p in self.coordinates if len(p) > 0]

    def get_points(self):
        ''' Enables the onclick events and returns the points.

        The format of the returned array:
        [array([[x1,...,xn],[y1,...,yn]]),array([[xn+1,...],[yn+1,...]]),...]
        Each 'array' element is a numpy.ndarray and represents one transect,
        where x1,y1 is the first point in the first transect,
        and xn,yn the last point in the first transect.
        The inner x/y-arrays are also numpy.ndarrays
        '''
        self.fig.canvas.mpl_connect('button_press_event', self.onclick)
        self.ax.set_xlim([0, self.data.shape[1]])
        self.ax.set_ylim([0, self.data.shape[0]])
        self.ax.invert_yaxis()
        self.ax.tick_params(direction='in', pad=-20, labelsize=10)

        self.text_ax.set_xlim([0, 1])
        self.text_ax.set_ylim([0, 1])
        self.text_ax.set_xticks([])
        self.text_ax.set_yticks([])

        text = ('To draw a transect line: click in several locations\n'
                'To start drawing a new line: press "space",'
                ' click in the next\n'
                '        location, release "space" and continue clicking\n'
                'To zoom: press "z" and use pan/zoom tools, then release "z"')
        self.text_ax.text(0.01, 0.9, text, fontsize=13,
                          verticalalignment='top', horizontalalignment='left')

        # collect data
        plt.show()

        # convert list of lists of coordinates to list of arrays
        points = self._convert_coordinates()

        return points



from nansat.mappers.mapper_arome import Mapper as MapperArome
from nansat.mappers.opendap import Opendap
from nansat.exceptions import WrongMapperError

class Mapper(Opendap, MapperArome):

    baseURLs = ['http://thredds.met.no/thredds/catalog/arome25/catalog.html']

    def __init__(self, *args, **kwargs):

        raise WrongMapperError('Mapper is under development...')

        fn = args[0]
        #ds = args[1] - None
        #mm = args[2] - None

        self.test_mapper(fn)



# Name:        mapper_modisL1
# Purpose:     Mapping for MODIS-L1 data
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from dateutil.parser import parse
import warnings
import json

import pythesint as pti

from nansat.tools import gdal, ogr
from nansat.exceptions import WrongMapperError
from nansat.vrt import VRT
from hdf4_mapper import HDF4Mapper


class Mapper(HDF4Mapper):
    ''' VRT with mapping of WKV for MODIS Level 1 (QKM, HKM, 1KM) '''

    def __init__(self, filename, gdalDataset, gdalMetadata, emrange='VNIR', **kwargs):
        ''' Create MODIS_L1 VRT '''
        # check mapper
        try:
            INSTRUMENTSHORTNAME = gdalMetadata['INSTRUMENTSHORTNAME']
        except:
            raise WrongMapperError
        if INSTRUMENTSHORTNAME != 'ASTER':
            raise WrongMapperError
        try:
            SHORTNAME = gdalMetadata['SHORTNAME']
        except:
            raise WrongMapperError
        if SHORTNAME != 'ASTL1B':
            raise WrongMapperError

        # set up metadict for data with various resolution
        subDSString = 'HDF4_EOS:EOS_SWATH:"%s":%s:%s'
        metaDictVNIR = [
        {'src': {'SourceFilename': subDSString % (filename, 'VNIR_Swath', 'ImageData1' )}, 'dst': {'wavelength': '560'}},
        {'src': {'SourceFilename': subDSString % (filename, 'VNIR_Swath', 'ImageData2' )}, 'dst': {'wavelength': '660'}},
        {'src': {'SourceFilename': subDSString % (filename, 'VNIR_Swath', 'ImageData3N')}, 'dst': {'wavelength': '820'}},
        {'src': {'SourceFilename': subDSString % (filename, 'VNIR_Swath', 'ImageData3B')}, 'dst': {'wavelength': '820'}},
        ]

        metaDictSWIR = [
        {'src': {'SourceFilename': subDSString % (filename, 'SWIR_Swath', 'ImageData4')}, 'dst': {'wavelength': '1650'}},
        {'src': {'SourceFilename': subDSString % (filename, 'SWIR_Swath', 'ImageData5')}, 'dst': {'wavelength': '2165'}},
        {'src': {'SourceFilename': subDSString % (filename, 'SWIR_Swath', 'ImageData6')}, 'dst': {'wavelength': '2205'}},
        {'src': {'SourceFilename': subDSString % (filename, 'SWIR_Swath', 'ImageData7')}, 'dst': {'wavelength': '2260'}},
        {'src': {'SourceFilename': subDSString % (filename, 'SWIR_Swath', 'ImageData8')}, 'dst': {'wavelength': '2330'}},
        {'src': {'SourceFilename': subDSString % (filename, 'SWIR_Swath', 'ImageData9')}, 'dst': {'wavelength': '2395'}},
        ]

        metaDictTIR = [
        {'src': {'SourceFilename': subDSString % (filename, 'TIR_Swath',  'ImageData10')}, 'dst': {'wavelength': '8300'}},
        {'src': {'SourceFilename': subDSString % (filename, 'TIR_Swath',  'ImageData11')}, 'dst': {'wavelength': '8650'}},
        {'src': {'SourceFilename': subDSString % (filename, 'TIR_Swath',  'ImageData12')}, 'dst': {'wavelength': '9100'}},
        {'src': {'SourceFilename': subDSString % (filename, 'TIR_Swath',  'ImageData13')}, 'dst': {'wavelength': '10600'}},
        {'src': {'SourceFilename': subDSString % (filename, 'TIR_Swath',  'ImageData14')}, 'dst': {'wavelength': '11300'}},
        ]

        # select appropriate metaDict based on <emrange> parameter
        metaDict = {'VNIR': metaDictVNIR,
                    'SWIR': metaDictSWIR,
                    'TIR': metaDictTIR,
                    }[emrange]

        # get 1st EOS subdataset and parse to VRT.__init__()
        # for retrieving geo-metadata
        try:
            gdalSubDataset0 = gdal.Open(metaDict[0]['src']['SourceFilename'])
        except (AttributeError, IndexError):
            raise WrongMapperError

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdalSubDataset0, metadata=gdalMetadata)

        # add source band, wkv and suffix
        for metaEntry in metaDict:
            metaEntry['src']['SourceBand'] = 1
            metaEntry['dst']['wkv'] = 'toa_outgoing_spectral_radiance'
            metaEntry['dst']['suffix'] = metaEntry['dst']['wavelength']

            if 'ImageData3N' in metaEntry['src']['SourceFilename']:
                metaEntry['dst']['suffix'] += 'N'

            if 'ImageData3B' in metaEntry['src']['SourceFilename']:
                metaEntry['dst']['suffix'] += 'B'

        # add scale and offset
        for metaEntry in metaDict:
            bandNo = metaEntry['src']['SourceFilename'].strip().split(':')[-1].replace('ImageData', '')
            metaEntry['src']['ScaleRatio'] = float(gdalMetadata['INCL' + bandNo])
            metaEntry['src']['ScaleOffset'] = float(gdalMetadata['OFFSET' + bandNo])

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        # set time
        datetimeString = self.find_metadata(gdalMetadata, "SETTINGTIMEOFPOINTING")
        # Adding valid time to dataset
        self.dataset.SetMetadataItem('time_coverage_start',
                                     parse(datetimeString+'+00').isoformat())
        self.dataset.SetMetadataItem('time_coverage_end',
                                     parse(datetimeString+'+00').isoformat())

        mm = pti.get_gcmd_instrument('ASTER')
        ee = pti.get_gcmd_platform('TERRA')
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))

        self._remove_geolocation()



# ------------------------------------------------------------------------------
# Name:         nansat_test_base.py
# Purpose:      Basic class for Nansat tests
#
# Author:       Anton Korosov
#
# Created:      20.03.2018
# Copyright:    (c) NERSC
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
# ------------------------------------------------------------------------------
from __future__ import print_function, absolute_import, division
import os
import sys
import unittest
import tempfile
import pythesint

from mock import patch, PropertyMock, Mock, MagicMock, DEFAULT

from nansat.tests import nansat_test_data as ntd


class NansatTestBase(unittest.TestCase):

    def setUp(self):
        self.test_file_gcps = os.path.join(ntd.test_data_path, 'gcps.tif')
        self.test_file_stere = os.path.join(ntd.test_data_path, 'stere.tif')
        self.test_file_complex = os.path.join(ntd.test_data_path, 'complex.nc')
        self.test_file_arctic = os.path.join(ntd.test_data_path, 'arctic.nc')
        self.tmp_data_path = ntd.tmp_data_path
        self.default_mapper = 'generic'
        fd, self.tmp_filename = tempfile.mkstemp(suffix='.nc', dir=ntd.tmp_data_path)
        os.close(fd)

        if not os.path.exists(self.test_file_gcps):
            raise ValueError('No test data available')
        # Mock several Pythesint functions to avoid network connection
        self.patcher = patch.multiple(pythesint, get_wkv_variable=DEFAULT,
                                                  get_gcmd_instrument=DEFAULT,
                                                  get_gcmd_platform=DEFAULT)
        self.mock_pti = self.patcher.start()
        self.mock_pti['get_gcmd_instrument'].return_value=dict(short_name='MODIS')
        self.mock_pti['get_gcmd_platform'].return_value=dict(short_name='AQUA')
        self.mock_pti['get_wkv_variable'].return_value=dict(short_name='swathmask')

    def tearDown(self):
        self.patcher.stop()
        try:
            os.unlink(self.tmp_filename)
        except OSError:
            pass




# Name:        nansat_mapper_merisL1
# Purpose:     Mapping for Meris-L1 data
# Authors:     Anton Korosov
# Licence:     This file is part of NANSAT. You can redistribute it or modify
#              under the terms of GNU General Public License, v.3
#              http://www.gnu.org/licenses/gpl-3.0.html
from dateutil.parser import parse
import json

from nansat.vrt import VRT
from nansat.exceptions import WrongMapperError
from envisat import Envisat

import pythesint as pti

class Mapper(VRT, Envisat):
    ''' VRT with mapping of WKV for MERIS Level 1 (FR or RR) '''

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 geolocation=False, zoomSize=500, step=1, **kwargs):

        ''' Create MER1 VRT

        Parameters
        -----------
        filename : string

        gdalDataset : gdal dataset

        gdalMetadata : gdal metadata

        geolocation : bool (default is False)
            if True, add gdal geolocation

        zoomSize: int (used in envisat.py)
            size, to which the ADS array will be zoomed using scipy
            array of this size will be stored in memory

        step: int (used in envisat.py)
            step of pixel and line in GeolocationArrays. lat/lon grids are
            generated at that step

        '''

        self.setup_ads_parameters(filename, gdalMetadata)

        if (self.product[0:9] != "MER_FRS_1" and
                self.product[0:9] != "MER_RR__1"):
            raise WrongMapperError

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdalDataset)

        metaDict = [{'src': {'SourceFilename': filename, 'SourceBand': 1},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '413'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 2},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '443'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 3},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '490'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 4},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '510'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 5},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '560'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 6},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '620'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 7},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '665'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 8},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '681'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 9},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '709'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 10},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '753'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 11},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '761'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 12},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '778'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 13},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '864'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 14},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '849'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 15},
                     'dst': {'wkv': 'toa_outgoing_spectral_radiance',
                             'wavelength': '900'}},
                    {'src': {'SourceFilename': filename, 'SourceBand': 16,
                             'DataType': 1},
                     'dst': {'wkv': 'quality_flags', 'suffix': 'l1'}}
                    ]

        # add DataType into 'src' and suffix into 'dst'
        for bandDict in metaDict:
            if 'DataType' not in bandDict['src']:
                # default for meris L1 DataType UInt16
                bandDict['src']['DataType'] = 2
            if 'wavelength' in bandDict['dst']:
                bandDict['dst']['suffix'] = bandDict['dst']['wavelength']

        # get scaling GADS from header
        scales = self.read_scaling_gads(range(7, 22))
        # set scale factor to the band metadata (only radiances)
        for i, bandDict in enumerate(metaDict[:-1]):
            bandDict['src']['ScaleRatio'] = str(scales[i])

        # get list with resized VRTs from ADS
        self.band_vrts = {'adsVRTs': self.get_ads_vrts(gdalDataset,
                                                     ['sun zenith angles',
                                                      'sun azimuth angles',
                                                      'zonal winds',
                                                      'meridional winds'],
                                                     zoomSize=zoomSize,
                                                     step=step)}
        # add bands from the ADS VRTs
        for adsVRT in self.band_vrts['adsVRTs']:
            metaDict.append({'src': {'SourceFilename': adsVRT.filename,
                                     'SourceBand': 1},
                             'dst': {'name': (adsVRT.dataset.GetRasterBand(1).
                                              GetMetadataItem('name')),
                                     'units': (adsVRT.dataset.GetRasterBand(1).
                                               GetMetadataItem('units'))}
                             })

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        # set time
        self._set_envisat_time(gdalMetadata)

        # Get dictionary describing the instrument and platform according to
        # the GCMD keywords
        mm = pti.get_gcmd_instrument('meris')
        ee = pti.get_gcmd_platform('envisat')

        # TODO: Validate that the found instrument and platform are indeed what we
        # want....

        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))

        # add geolocation arrays
        if geolocation:
            self.add_geolocation_from_ads(gdalDataset,
                                          zoomSize=zoomSize, step=step)



# Name:    geolocation.py
# Purpose: Container of Geolocation class
# Authors:      Anton Korosov
# Created:      14.01.2018
# Copyright:    (c) NERSC 2011 - 2018
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
from __future__ import absolute_import

import gdal, osr

from nansat.nsr import NSR


class Geolocation(object):
    """Container for GEOLOCATION data

    Keeps references to bands with X and Y coordinates, offset and step
    of pixel and line. All information is stored in dictionary self.data

    Instance of Geolocation is used in VRT and ususaly created in
    a Mapper.
    """
    # instance attributes
    data = None
    x_vrt = None
    y_vrt = None

    def __init__(self, x_vrt, y_vrt, **kwargs):
        """Create Geolocation object from input VRT objects

        Parameters
        -----------
        x_vrt_name : VRT
            dataset with X-coordinates
        y_vrt_name : VRT
            dataset with Y-coordinates
        **kwargs : dict
            parameters for self._init_data()

        Note
        ----
        Modifies, self.x_vrt, self.y_vrt, self.data

        """
        # dictionary with all metadata
        self.data = dict()
        # VRT objects
        self.x_vrt = x_vrt
        self.y_vrt = y_vrt

        self._init_data(x_vrt.filename, y_vrt.filename, **kwargs)

    def _init_data(self, x_filename, y_filename, x_band=1, y_band=1, srs=NSR().wkt,
                                                line_offset=0, line_step=1,
                                                pixel_offset=0, pixel_step=1):
        """Init data of Geolocation object from input parameters
        Parameters
        -----------
        x_filename : str
            name of file for X-dataset
        y_filename : str
            name of file for Y-dataset
        x_band : int
            number of the band in the X-dataset
        y_band : int
            number of the band in the Y-dataset
        srs : str
            WKT
        line_offset : int
            offset of first line
        line_step : int
            step of lines
        pixel_offset : int
            offset of first pixel
        pixel_step : int
            step of pixels

        Notes
        -----
        Saves everything in self.data dict

        """
        self.data['SRS'] = srs
        self.data['X_DATASET'] = x_filename
        self.data['Y_DATASET'] = y_filename
        self.data['X_BAND'] = str(x_band)
        self.data['Y_BAND'] = str(y_band)
        self.data['LINE_OFFSET'] = str(line_offset)
        self.data['LINE_STEP'] = str(line_step)
        self.data['PIXEL_OFFSET'] = str(pixel_offset)
        self.data['PIXEL_STEP'] = str(pixel_step)

    @classmethod
    def from_dataset(cls, dataset):
        """Create geolocation from GDAL dataset
        Parameters
        ----------
        dataset : gdal.Dataset
            input dataset to copy Geolocation metadata from
        """
        self = cls.__new__(cls) # empty object
        self.x_vrt = None
        self.y_vrt = None
        self.data = dataset.GetMetadata('GEOLOCATION')
        return self

    @classmethod
    def from_filenames(cls, x_filename, y_filename, **kwargs):
        """Create geolocation from names of files with geolocation
        Parameters
        ----------
        x_filename : str
            name of file for X-dataset
        y_filename : str
            name of file for Y-dataset
        **kwargs : dict
            parameters for self._init_data()
        """
        self = cls.__new__(cls) # empty object
        self.x_vrt = None
        self.y_vrt = None
        self.data = {}
        self._init_data(x_filename, y_filename, **kwargs)
        return self

    def get_geolocation_grids(self):
        """Read values of geolocation grids"""
        lon_dataset = gdal.Open(self.data['X_DATASET'])
        lon_grid = lon_dataset.GetRasterBand(int(self.data['X_BAND'])).ReadAsArray()
        lat_dataset = gdal.Open(self.data['Y_DATASET'])
        lat_grid = lat_dataset.GetRasterBand(int(self.data['Y_BAND'])).ReadAsArray()
        return lon_grid, lat_grid



# Name:         mapper_ncep
# Purpose:      Nansat mapping for NCEP GFS model data
# Author:       Knut-Frode Dagestad
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
#
# Made for GRIB files downloaded from http://nomads.ncep.noaa.gov/
# NB: Band numbers is hardcoded for band subsets extracted at NERSC,
# mapper will not work for other NCEP GFS files before made more generic
import datetime
import json
import numpy as np
from dateutil.parser import parse
import pythesint as pti

from nansat.vrt import VRT
from nansat.exceptions import WrongMapperError


class Mapper(VRT):
    ''' VRT with mapping of WKV for NCEP GFS '''

    def __init__(self, filename, gdalDataset, gdalMetadata, **kwargs):
        ''' Create NCEP VRT '''

        if not gdalDataset:
            raise WrongMapperError(filename)

        geotransform = gdalDataset.GetGeoTransform()
        if (geotransform == (-0.25, 0.5, 0.0, 90.25, 0.0, -0.5) or
                geotransform == (-0.5, 1.0, 0.0, 90.5, 0.0, -1.0)):
            if gdalDataset.RasterCount == 4:
                srcBandId = {'temperature': 2,
                             'u-component': 3,
                             'v-component': 4}
            elif gdalDataset.RasterCount == 9:
                srcBandId = {'temperature': 6,
                             'u-component': 8,
                             'v-component': 9}
            else:
                raise WrongMapperError(filename)
        else:
            raise WrongMapperError(filename)  # Not water proof

        # Adding valid time from the GRIB file to dataset
        band = gdalDataset.GetRasterBand(srcBandId['u-component'])
        validTime = band.GetMetadata()['GRIB_VALID_TIME']
        time_isoformat = (datetime.datetime.utcfromtimestamp(
            int(validTime.strip().split(' ')[0])).isoformat())

        # Set band metadata time_iso_8601 for use in OpenWind
        time_iso_8601 = np.datetime64(parse(time_isoformat))
        metaDict = [{'src': {'SourceFilename': filename,
                             'SourceBand': srcBandId['u-component']},
                     'dst': {'wkv': 'eastward_wind',
                             'height': '10 m',
                             'time_iso_8601': time_iso_8601}},
                    {'src': {'SourceFilename': filename,
                             'SourceBand': srcBandId['v-component']},
                     'dst': {'wkv': 'northward_wind',
                             'height': '10 m',
                             'time_iso_8601': time_iso_8601}},
                    {'src': [{'SourceFilename': filename,
                              'SourceBand': srcBandId['u-component'],
                              'DataType': (gdalDataset.GetRasterBand(srcBandId['u-component']).DataType)
                              },
                             {'SourceFilename': filename,
                              'SourceBand': srcBandId['v-component'],
                              'DataType': gdalDataset.GetRasterBand(srcBandId['v-component']).DataType
                              }],
                     'dst': {'wkv': 'wind_speed',
                             'PixelFunctionType': 'UVToMagnitude',
                             'name': 'windspeed',
                             'height': '2 m',
                             'time_iso_8601': time_iso_8601
                             }},
                    {'src': [{'SourceFilename': filename,
                              'SourceBand': srcBandId['u-component'],
                              'DataType': gdalDataset.GetRasterBand(srcBandId['u-component']).DataType
                              },
                             {'SourceFilename': filename,
                              'SourceBand': srcBandId['v-component'],
                              'DataType': gdalDataset.GetRasterBand(srcBandId['v-component']).DataType
                              }],
                     'dst': {'wkv': 'wind_from_direction',
                             'PixelFunctionType': 'UVToDirectionFrom',
                             'name': 'winddirection',
                             'height': '2 m',
                             'time_iso_8601': time_iso_8601
                             }},
                    {'src': {'SourceFilename': filename,
                             'SourceBand': srcBandId['temperature']},
                     'dst': {'wkv': 'air_temperature',
                             'name': 'air_t',
                             'height': '2 m',
                             'time_iso_8601': time_iso_8601}
                     }]

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(gdalDataset)

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        self.dataset.SetMetadataItem('time_coverage_start', time_isoformat)

        self.dataset.SetMetadataItem('time_coverage_end', time_isoformat)

        # Get dictionary describing the instrument and platform according to
        # the GCMD keywords
        mm = pti.get_gcmd_instrument('computer')
        ee = pti.get_gcmd_platform('ncep-gfs')
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))




# Name:    nansat_map.py
# Purpose: Container of Nansatap class
# Authors:      Asuka Yamakawa, Anton Korosov, Knut-Frode Dagestad,
#               Morten W. Hansen, Alexander Myasoyedov,
#               Dmitry Petrenko, Evgeny Morozov
# Created:      29.06.2011
# Copyright:    (c) NERSC 2011 - 2013
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
from __future__ import absolute_import

import os
import re
import warnings

import numpy as np

try:
    if 'DISPLAY' not in os.environ:
        import matplotlib; matplotlib.use('Agg')
    from mpl_toolkits.basemap import Basemap
    import matplotlib as mpl
    from matplotlib import cm
    import matplotlib.pyplot as plt
    from scipy import ndimage
except ImportError:
    BASEMAP_LIB_IS_INSTALLED = False
    MATPLOTLIB_IS_INSTALLED = False
else:
    BASEMAP_LIB_IS_INSTALLED = True
    MATPLOTLIB_IS_INSTALLED = True


from nansat.nsr import NSR
from nansat.tools import get_random_color
from nansat.warnings import NansatFutureWarning

NANSATMAP_WARNING = ('Nansatmap() will be disabled in Nansat 1.1. and moved to a separate package'
                     'It is not covered by unittests intentionally.')
warnings.warn(NANSATMAP_WARNING, NansatFutureWarning)

class Nansatmap(Basemap):
    '''Perform opeartions with graphical files: create,
    add legend and geolocation_grids, save.

    NansatMap instance is created in the Nansat.write_map method.
    The methods below are applied consequently in order to get projection,
    generate a basemap from array(s), add legend and geolocation grids,
    save to a file.

    '''
    # general attributes
    cmap = cm.jet
    colorbar = None
    mpl = []
    lon, lat, x, y = None, None, None, None
    # parameters for smoothing
    # convolve
    convolve_weightSize = 7
    convolve_weights = None
    convolve_mode = 'reflect'
    convolve_cval = 0.0
    convolve_origin = 0
    # fourier_gaussian
    fourier_sigma = 1.0
    fourier_n = -1
    fourier_axis = -1
    # spline
    spline_order = 3
    spline_axis = -1
    # gaussian filter
    gaussian_sigma = 2.5
    gaussian_order = 0
    gaussian_mode = 'reflect'
    gaussian_cval = 0.0
    # saving parameters
    DEFAULT_EXTENSION = '.png'


    def __init__(self, domain, **kwargs):
        ''' Set attributes
        Get proj4 from the given domain and convert the proj4 projection to
        the basemap projection.

        Parameters
        -----------
        domain : domain object
        kwargs : dictionary
            parameters that are used for all operations.

        Modifies
        ---------
        self.fig : figure
            matplotlib.pyplot.figure
        self.colorbar : boolean
            if colorbar is True, it is possible to put colorbar.
            e.g. contour_plots(contour_style='fill'), put_color()
        self.mpl : list
            elements are matplotlib.contour.QuadContourSet instance,
                         matplotlib.quiver.Quiver instance or
                         matplotlib.collections.QuadMesh object

        See also
        ----------
        http://matplotlib.org/basemap/api/basemap_api.html

        '''
        if not MATPLOTLIB_IS_INSTALLED:
            raise ImportError('Matplotlib is not installed')
        if not BASEMAP_LIB_IS_INSTALLED:
            raise ImportError('Basemap is not installed')
        warnings.warn(NANSATMAP_WARNING, NansatFutureWarning)

        self.domain = domain

        # get proj4
        proj4 = NSR(domain.vrt.get_projection()).ExportToProj4()

        # convert proj4 to basemap projection
        projStr = proj4.split(' ')[0][6:]
        projection = {'aea': 'aea', 'ocea': 'aea',
                      'aeqd': 'aeqd', 'xxx1': 'spaeqd', 'xxx2': 'npaeqd',
                      'cass': 'cass',
                      'cea': 'cea',
                      'eqc': 'cyl', 'longlat': 'cyl',
                      'eck4': 'eck4',
                      'eqdc': 'eqdc',
                      'gall': 'gall',
                      'geos': 'geos',
                      'gnom': 'gnom',
                      'hammer': 'hammer', 'nell_h': 'hammer',
                      'kav7': 'kav7',
                      'laea': 'laea', 'xxx3': 'splaea', 'xxx4': 'nplaea',
                      'lcc': 'lcc', 'lcca': 'lcc',
                      'mbtfpq': 'mbtfpq',
                      'somerc': 'merc', 'merc': 'merc', 'omerc': 'merc',
                      'mill': 'mill',
                      'moll': 'moll',
                      'nsper': 'nsper',
                      'omerc': 'omerc',
                      'ortho': 'ortho',
                      'poly': 'poly', 'rpoly': 'poly', 'imw_p': 'poly',
                      'robin': 'robin',
                      'sinu': 'sinu', 'fouc_s': 'sinu', 'gn_sinu': 'sinu',
                      'mbtfps': 'sinu', 'urmfps': 'sinu',
                      'stere': 'stere', 'sterea': 'stere', 'lee_os': 'stere',
                      'mil_os': 'stere', 'rouss': 'stere',
                      'ups': 'npstere', 'ups': 'spstere',  # CHECK!!
                      'tmerc': 'tmerc', 'gstmerc': 'tmerc', 'utm': 'tmerc',
                      'vandg': 'vandg', 'vandg2': 'vandg',
                      'vandg3': 'vandg', 'vandg4': 'vandg',
                      }.get(projStr, 'cyl')

        if projection in ['stere']:
            lon_0 = float(re.findall('lon_0=+[-+]?\d*[.\d*]*',
                                     proj4)[0].split('=')[1])
            lat_0 = float(re.findall('lat_0=+[-+]?\d*[.\d*]*',
                                     proj4)[0].split('=')[1])
            kwargs['lon_0'] = lon_0
            kwargs['lat_0'] = lat_0

        if projStr == 'utm':
            kwargs['lon_0'] = -180 + NSR(proj4).GetUTMZone()*6 - 3
            kwargs['lat_0'] = 0

        self.extensionList = ['png', 'emf', 'eps', 'pdf', 'rgba',
                              'ps', 'raw', 'svg', 'svgz']

        # set llcrnrlat, urcrnrlat, llcrnrlon and urcrnrlon to kwargs.
        # if required, modify them from -90. to 90.
        # get min/max lon/lat
        lonCrn, latCrn = domain.get_corners()
        self.lonMin = min(lonCrn)
        self.lonMax = max(lonCrn)
        self.latMin = max(min(latCrn), -90.)
        self.latMax = min(max(latCrn), 90.)

        if not('llcrnrlat' in kwargs.keys()):
            kwargs['llcrnrlat'] = latCrn[1]
        if not('urcrnrlat' in kwargs.keys()):
            kwargs['urcrnrlat'] = latCrn[2]
        if not('llcrnrlon' in kwargs.keys()):
            kwargs['llcrnrlon'] = lonCrn[1]
        if not('urcrnrlon' in kwargs.keys()):
            kwargs['urcrnrlon'] = lonCrn[2]

        # separate kwarge of plt.figure() from kwargs
        figArgs = ['num', 'figsize', 'dpi', 'facecolor', 'edgecolor',
                   'frameon']
        figKwargs = {}
        for iArg in figArgs:
            if iArg in kwargs.keys():
                figKwargs[iArg] = kwargs.pop(iArg)

        Basemap.__init__(self, projection=projection, **kwargs)

        # create figure and set it as an attribute
        plt.close()
        self.fig = plt.figure(**figKwargs)

    def smooth(self, idata, mode, **kwargs):
        '''Smooth data for contour() and contourf()

        idata is smoothed by convolve, fourier_gaussian, spline or
        gaussian (default). If contour_mode is 'convolve' and weight is None,
        the weight matrix is created automatically.

        Parameters
        -----------
        idata : numpy 2D array
            Input data
        mode : string
            'convolve','fourier','spline' or 'gaussian'

        Returns
        ---------
        odata : numpy 2D array

        See also
        ----------
        http://docs.scipy.org/doc/scipy/reference/ndimage.html

        '''
        # modify default parameter
        self._set_defaults(kwargs)

        if mode == 'convolve':
            # if weight is None, create a weight matrix
            if self.convolve_weights is None:
                weights = np.ones((self.convolve_weightSize,
                                   self.convolve_weightSize))
                center = (self.convolve_weightSize - 1) / 2
                for i in range(- (center), center + 1, 1):
                    for j in range(- (center), center + 1, 1):
                        weights[i][j] /= pow(2.0, max(abs(i), abs(j)))
                self.convolve_weights = weights
            odata = ndimage.convolve(idata,
                                     weights=self.convolve_weights,
                                     mode=self.convolve_mode,
                                     cval=self.convolve_cval,
                                     origin=self.convolve_origin)
        elif mode == 'fourier':
            odata = ndimage.fourier_gaussian(idata,
                                             sigma=self.fourier_sigma,
                                             n=self.fourier_n,
                                             axis=self.fourier_axis)
        elif mode == 'spline':
            odata = ndimage.spline_filter1d(idata,
                                            order=self.spline_order,
                                            axis=self.spline_axis)
        else:
            if mode != 'gaussian':
                print('apply Gaussian filter in image_process()')
            odata = ndimage.gaussian_filter(idata,
                                            sigma=self.gaussian_sigma,
                                            order=self.gaussian_order,
                                            mode=self.gaussian_mode,
                                            cval=self.gaussian_cval)
        return odata

    def _do_contour(self, bmfunc, data, v, smooth, mode, **kwargs):
        ''' Prepare data and make contour or contourf plots

        1. Smooth data
        1. Add colormap
        1. Append contour or contourf plot to self.mpl

        bmfunc : Basemap function
            Basemap.contour, Basemap.contourf
        data : numpy 2D array
            Input data
        v : list with values
            draw contour lines at the values specified in sequence v
        smooth : Boolean
            Apply smoothing?
        mode : string
            'gaussian', 'spline', 'fourier', 'convolve'
            mname of smoothing algorithm to apply

        '''
        self._create_xy_grids()

        # if cmap is given, set to self.cmap
        if 'cmap' in kwargs.keys():
            self.cmap = kwargs.pop('cmap')

        # smooth data
        if smooth:
            data = self.smooth(data, mode, **kwargs)

        # draw contour lines
        if v is None:
            self.mpl.append(bmfunc(self, self.x, self.y, data, **kwargs))
        else:
            self.mpl.append(bmfunc(self, self.x, self.y, data, v, **kwargs))

    def contour(self, data, v=None, smooth=False, mode='gaussian',
                label=True, **kwargs):
        '''Draw lined contour plots

        If smooth is True, data is smoothed. Then draw lined contour.

        Parameters
        ----------
        data : numpy 2D array
            Input data
        v : list with values
            draw contour lines at the values specified in sequence v
        smooth : Boolean
            Apply smoothing?
        mode : string
            'gaussian', 'spline', 'fourier', 'convolve'
            mname of smoothing algorithm to apply
        label : boolean
            Add lables?
        **kwargs:
            Optional parameters for Nansatmap.smooth()
            Optional parameters for pyplot.contour().
            Optional parameters for pyplot.clabel()

        Modifies
        ---------
        self.mpl : list
            append QuadContourSet instance
        '''

        self._do_contour(Basemap.contour, data, v, smooth, mode, **kwargs)

        # add lables to the contour lines
        if label:
            plt.clabel(self.mpl[-1], **kwargs)

    def contourf(self, data, v=None,
                 smooth=False, mode='gaussian', **kwargs):
        '''Draw filled contour plots

        If smooth is True, data is smoothed. Then draw filled contour.

        Parameters
        ----------
        data : numpy 2D array
            Input data
        v : list with values
            draw contour lines at the values specified in sequence v
        smooth : Boolean
            Apply smoothing?
        mode : string
            'gaussian', 'spline', 'fourier', 'convolve'
            mname of smoothing algorithm to apply
        **kwargs:
            cmap : colormap (e.g. cm.jet)
            Optional parameters for Nansatmap.smooth()
            Optional parameters for pyplot.contourf().

        Modifies
        ---------
        self.mpl : list
            append QuadContourSet instance

        '''
        self._do_contour(Basemap.contourf, data, v, smooth, mode, **kwargs)
        self.colorbar = len(self.mpl) - 1

    def imshow(self, data, low=0, high=255, **kwargs):
        ''' Make RGB plot over the map

        data : numpy array
            RGB or RGBA input data
        **kwargs:
            Parameters for Basemap.imshow

        Modifies
        ---------
        self.mpl : list
            append AxesImage object with imshow

        '''
        # Create X/Y axes
        self._create_xy_grids()

        # add random colormap
        if 'cmap' in kwargs and kwargs['cmap'] == 'random':
            values = np.unique(data[np.isfinite(data)])
            cmap, norm = self._create_random_colormap(values,
                                                      low=low, high=high)
            kwargs['cmap'] = cmap
            kwargs['norm'] = norm

        # Plot data using imshow
        self.mpl.append(Basemap.imshow(self, data,
                                       extent=[self.x.min(), self.x.max(),
                                               self.y.min(), self.y.max()],
                                       origin='upper', **kwargs))
        self.colorbar = len(self.mpl) - 1

    def pcolormesh(self, data, **kwargs):
        '''Make a pseudo-color plot over the map

        Parameters
        ----------
        data : numpy 2D array
            Input data
        **kwargs:
            Parameters for Basemap.pcolormesh (e.g. vmin, vmax)

        Modifies
        ---------
        self.mpl : list
            append matplotlib.collections.QuadMesh object

        '''
        # mask nan data
        data = np.ma.array(data, mask=np.isnan(data))
        # Plot a quadrilateral mesh.
        self._create_xy_grids()
        self.mpl.append(Basemap.pcolormesh(self, self.x, self.y, data,
                                           **kwargs))
        self.colorbar = len(self.mpl) - 1

    def quiver(self, dataX, dataY, step=None, quivectors=None, **kwargs):
        '''Draw quiver plots

        Parameters
        ----------
        dataX :  numpy array
            Input data with X-component
        dataY :  numpy array
            Input data with Y-component
        step : int or (int, int)
            Skip <step> pixels along both dimentions(alternative to quivectors)
        quivectors : int or (int,int)
            Number of vectors along both dimentions
        Parameters for Basemap.quiver()

        Modifies
        ---------
        self.mpl : list
            append matplotlib.quiver.Quiver instance

        '''
        # if Nan is included, apply mask
        dataX = np.ma.array(dataX, mask=np.isnan(dataX))
        dataY = np.ma.array(dataY, mask=np.isnan(dataY))

        # get subsetting parameters
        if type(step) is int:
            step0 = step1 = step
        elif type(step) in [list, tuple]:
            step0 = step[0]
            step1 = step[1]
        elif quivectors is not None:
            if type(quivectors) is int:
                quivectors0 = quivectors
                quivectors1 = quivectors
            if type(quivectors) in [list, tuple]:
                quivectors0 = quivectors[0]
                quivectors1 = quivectors[1]
            step0 = dataX.shape[0] / quivectors0
            step1 = dataX.shape[1] / quivectors1
        else:
            step0 = step1 = 5

        dataX2 = dataX[::step0, ::step1]
        dataY2 = dataY[::step0, ::step1]
        self._create_lonlat_grids()
        lon2 = self.lon[::step0, ::step1]
        lat2 = self.lat[::step0, ::step1]
        x2, y2 = self(lon2, lat2)

        qKwargs = {}
        for iKey in ['width', 'scale', 'units', 'angles', 'scale_units']:
            if iKey in kwargs.keys():
                qKwargs[iKey] = kwargs.pop(iKey)
        Q = Basemap.quiver(self, x2, y2, dataX2, dataY2, **qKwargs)

        qkargs = {}
        for iKey in ['X', 'Y', 'U', 'label']:
            if iKey in kwargs.keys():
                qkargs[iKey] = kwargs.pop(iKey)

        if all(iKey in qkargs.keys() for iKey in ('X', 'Y', 'U', 'label')):
            self.mpl.append(plt.quiverkey(Q, qkargs['X'], qkargs['Y'],
                                          qkargs['U'], qkargs['label'],
                                          **kwargs))
        else:
            self.mpl.append(Q)

    def add_colorbar(self, fontsize=6, **kwargs):
        '''Add color bar

        Parameters
        ----------
        fontsize : int
        Parameters for matplotlib.pyplot.colorbar

        Modifies
        ---------
        Adds colorbar to self.fig

        '''
        if kwargs is None:
            kwargs = {}
        if not ('orientation' in kwargs.keys()):
            kwargs['orientation'] = 'horizontal'
        if not ('pad' in kwargs.keys()):
            kwargs['pad'] = 0.01

        # add colorbar and set font size
        if self.colorbar is not None:
            origin = self.mpl[self.colorbar]

            # if colormap is ListedColormap
            # add integer ticks
            ticks = None
            listedColormap = False
            if (hasattr(origin, 'cmap') and
                (type(origin.cmap) == mpl.colors.ListedColormap) and
                hasattr(origin.norm, 'boundaries')) :
                ticks = (origin.norm.boundaries[:-1] +
                         np.diff(origin.norm.boundaries) / 2.)
                listedColormap = True

            cbar = self.fig.colorbar(origin, ticks=ticks, **kwargs)
            if listedColormap:
                labels = origin.norm.boundaries[:-1]
                if np.all(labels == np.floor(labels)):
                    labels = labels.astype('int32')
                cbar.ax.set_xticklabels(labels)
            imaxes = plt.gca()
            plt.axes(cbar.ax)
            plt.xticks(fontsize=fontsize)
            plt.axes(imaxes)

    def drawgrid(self, lat_num=5, lon_num=5,
                 lat_labels=[True, False, False, False],
                 lon_labels=[False, False, True, False],
                 **kwargs):
        '''Draw and label parallels (lat and lon lines) for values (in degrees)

        Parameters
        -----------
        fontsize : int
        lat_num : int
            Number of latitude lables
        lon_num :
            Number of longitude lables
        lat_labels : list of Bool
            Location of latitude labels
        lon_labels : list of Bool
            Location of longitude labels

        See also: Basemap.drawparallels(), Basemap.drawmeridians()

        '''
        self.drawparallels(np.arange(self.latMin, self.latMax,
                           (self.latMax - self.latMin) / lat_num),
                           labels=lat_labels, **kwargs)
        self.drawmeridians(np.arange(self.lonMin, self.lonMax,
                           (self.lonMax - self.lonMin) / lon_num),
                           labels=lon_labels, **kwargs)

    def draw_continents(self, **kwargs):
        ''' Draw continents

        Parameters
        ----------
        Parameters for basemap.fillcontinents

        '''

        if kwargs is None:
            kwargs = {}
        if not ('color' in kwargs.keys()):
            kwargs['color'] = '#999999'
        if not ('lake_color' in kwargs.keys()):
            kwargs['lake_color'] = '#99ffff'

        # draw continets
        self.fillcontinents(**kwargs)

    def save(self, fileName, landmask=True, dpi=75,
             pad_inches=0, bbox_inches='tight', **kwargs):
        '''Draw continents and save

        Parameters
        -----------
        fileName : string
            name of outputfile
        landmask : Boolean
            Draw landmask?
        Parameters for basemap.fillcontinents

        '''
        if landmask:
            self.draw_continents(**kwargs)

        # set default extension
        if not((fileName.split('.')[-1] in self.extensionList)):
            fileName = fileName + self.DEFAULT_EXTENSION
        self.fig.savefig(fileName, dpi=dpi,
                         pad_inches=pad_inches,
                         bbox_inches=bbox_inches)

    def _set_defaults(self, idict):
        '''Check input params and set defaut values

        Look throught default parameters (self.d) and given parameters (dict)
        and paste value from input if the key matches

        Parameters
        ----------
        idict : dictionary
            parameter names and values

        Modifies
        ---------
            default self attributes

        '''
        for key in idict:
            if hasattr(self, key):
                setattr(self, key, idict[key])

    def _create_lonlat_grids(self):
        '''Generate grids with lon/lat coordinates in each cell

        Modifies
        ---------
        self.lon : numpy array with lon coordinates
        self.lat : numpy array with lat coordinates
        '''
        if self.lon is None or self.lat is None:
            self.lon, self.lat = self.domain.get_geolocation_grids()

    def _create_xy_grids(self):
        '''Generate grids with x/y coordinates in each cell

        Modifies
        ---------
        self.x : numpy array with X coordinates
        self.y : numpy array with Y coordinates
        '''
        self._create_lonlat_grids()
        if self.x is None or self.y is None:
            self.x, self.y = self(self.lon, self.lat)

    def _create_random_colormap(self, values, low=0, high=255):
        ''' Generate colormap and colorbar with random discrete colors

        Parameters
        ----------
            values : list or 1D array
                values for which the random colors are to be generated
        Returns
        -------
            cmap : matplotlib.color.Colormap
            norm : matplotlib.color.BoundaryNorm
        '''
        # create first random color
        randomColors = [get_random_color(low=low, high=high)]
        # add more random colors
        for v in values[1:]:
            randomColors.append(get_random_color(randomColors[-1],
                                                 low=low, high=high))

        # create colormap and norm
        cmap = mpl.colors.ListedColormap(randomColors)
        bounds = sorted(list(values))
        bounds += [max(bounds) + 1]  # bounds should be longer than values by 1
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

        return cmap, norm

    def add_zone_labels(self, zones, fontsize=5):
        ''' Finds best place of labels for a zone map, adds labels to the map

        Parameters
        ----------
            zones : numpy array with integer zones
                the same array as usied in Nansatmap.imshow
        '''
        zoneIndices = np.unique(zones[np.isfinite(zones)])
        for zi in zoneIndices:
            zrows, zcols = np.nonzero(zones == zi)
            zrc = np.median(zrows)
            zcc = np.median(zcols)
            lon, lat = self.domain.transform_points([zcc], [zrc], 0)
            x, y = self(lon[0], lat[0])
            plt.text(x, y, '%d' % zi, fontsize=fontsize)



# Name:        mapper_obpg_l2
# Purpose:     Mapping for L2 data from the OBPG web-site
# Authors:      Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import os
from datetime import datetime, timedelta
from math import ceil
import json

import pythesint as pti

from nansat.tools import gdal, ogr, parse_time
from nansat.exceptions import WrongMapperError
from nansat.vrt import VRT
from nansat.nsr import NSR


class Mapper(VRT):
    ''' Mapper for AMSR-2 L1 data

    '''

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 GCP_STEP=20, MAX_LAT=90, MIN_LAT=50, resolution='low',
                 **kwargs):
        ''' Create VRT
        Parameters
        ----------
        GCP_COUNT : int
            number of GCPs along each dimention
        '''
        ifile = os.path.split(filename)[1]
        if not ifile.startswith('GW1AM2_') or not ifile.endswith('.h5'):
            raise WrongMapperError
        try:
            ProductName = gdalMetadata['ProductName']
            PlatformShortName = gdalMetadata['PlatformShortName']
            SensorShortName = gdalMetadata['SensorShortName']
        except:
            raise WrongMapperError

        if (not ProductName == 'AMSR2-L1R' or
            not PlatformShortName == 'GCOM-W1' or
            not SensorShortName == 'AMSR2'):
            raise WrongMapperError

        if resolution == 'low':
            subDatasetWidth = 243
        else:
            subDatasetWidth = 486

        # get GCPs from lon/lat grids
        latGrid = gdal.Open('HDF5:"%s"://Latitude_of_Observation_Point_for_89A' % filename).ReadAsArray()
        lonGrid = gdal.Open('HDF5:"%s"://Longitude_of_Observation_Point_for_89A' % filename).ReadAsArray()
        if subDatasetWidth == 243:
            latGrid = latGrid[:, ::2]
            lonGrid = lonGrid[:, ::2]

        dx = .5
        dy = .5
        gcps = []
        k = 0
        maxY = 0
        minY = latGrid.shape[0]
        for i0 in range(0, latGrid.shape[0], GCP_STEP):
            for i1 in range(0, latGrid.shape[1], GCP_STEP):
                # create GCP with X,Y,pixel,line from lat/lon matrices
                lon = float(lonGrid[i0, i1])
                lat = float(latGrid[i0, i1])
                if (lon >= -180 and
                    lon <= 180 and
                    lat >= MIN_LAT and
                    lat <= MAX_LAT):
                    gcp = gdal.GCP(lon, lat, 0, i1 + dx, i0 + dy)
                    gcps.append(gcp)
                    k += 1
                    maxY = max(maxY, i0)
                    minY = min(minY, i0)
        yOff = minY
        ySize = maxY - minY

        # remove Y-offset from gcps
        for gcp in gcps:
            gcp.GCPLine -= yOff

        metaDict = []

        subDatasets = gdalDataset.GetSubDatasets()
        metadata = gdalDataset.GetMetadata()
        for subDataset in subDatasets:
            # select subdatasets fro that resolution (width)
            if (subDatasetWidth == int(subDataset[1].split(']')[0].split('x')[-1]) and
                'Latitude' not in subDataset[0] and 'Longitude' not in subDataset[0]):
                name = subDataset[0].split('/')[-1]
                # find scale
                scale = 1
                for meta in metadata:
                    if name + '_SCALE' in meta:
                        scale = float(metadata[meta])
                # create meta entry
                metaEntry = {'src': {'SourceFilename': subDataset[0],
                                     'sourceBand':  1,
                                     'ScaleRatio': scale,
                                     'ScaleOffset': 0,
                                     'yOff': yOff,
                                     'ySize': ySize,},
                             'dst': {'name': name}
                             }
                metaDict.append(metaEntry)

        # create VRT from one of the subdatasets
        gdalSubDataset = gdal.Open(metaEntry['src']['SourceFilename'])
        self._init_from_dataset_params(subDatasetWidth, ySize, (1,0,0,ySize,0,-1), NSR().wkt)
        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        self.dataset.SetMetadataItem('time_coverage_start',
                    parse_time(gdalMetadata['ObservationStartDateTime']).isoformat())
        self.dataset.SetMetadataItem('time_coverage_end',
                    parse_time(gdalMetadata['ObservationEndDateTime']).isoformat())
        # append GCPs and lat/lon projection to the vsiDataset
        self.dataset.SetGCPs(gcps, NSR().wkt)
        self.reproject_GCPs('+proj=stere +datum=WGS84 +ellps=WGS84 +lat_0=90 +lon_0=0 +no_defs')
        self.tps = True

        mm = pti.get_gcmd_instrument('AMSR2')
        ee = pti.get_gcmd_platform('GCOM-W1')
        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))



# Name:        mapper_ascat_nasa
# Purpose:     Mapping for ASCAT scatterometer winds
# Authors:     Knut-Frode Dagestad, Morten W. Hansen
# Licence:     This file is part of NANSAT. You can redistribute it or modify
#              under the terms of GNU General Public License, v.3
#              http://www.gnu.org/licenses/gpl-3.0.html
#
# For NetCDF files of ASCAT wind data from the NASA JPL archive:
# ftp://podaac-ftp.jpl.nasa.gov/allData/ascat/preview/L2/metop_a/12km/
import os.path
import datetime
import warnings

import json
import pythesint as pti

from nansat.tools import gdal, ogr
from nansat.geolocation import Geolocation
from nansat.vrt import VRT
from nansat.exceptions import WrongMapperError


class Mapper(VRT):
    ''' Create VRT with mapping of WKV '''

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 latlonGrid=None, mask='', **kwargs):

        ''' Create VRT

        Parameters
        -----------
        filename : string
        gdalDataset : gdal dataset
        gdalMetadata : gdal metadata
        latlonGrid : numpy 2 layered 2D array with lat/lons of desired grid
        '''
        # test if input files is ASCAT
        iDir, iFile = os.path.split(filename)
        iFileName, iFileExt = os.path.splitext(iFile)
        try:
            assert iFileName[0:6] == 'ascat_' and iFileExt == '.nc'
        except:
            raise WrongMapperError

        # Create geolocation
        subDataset = gdal.Open('NETCDF:"' + filename + '":lat')
        self.GeolocVRT = VRT(subDataset.RasterXSize, subDataset.RasterYSize)

        GeolocMetaDict = [{'src': {'SourceFilename': ('NETCDF:"' + filename +
                                                      '":lon'),
                                   'SourceBand': 1,
                                   'ScaleRatio': 0.00001,
                                   'ScaleOffset': -360},
                           'dst': {}},
                          {'src': {'SourceFilename': ('NETCDF:"' + filename +
                                                      '":lat'),
                                   'SourceBand': 1,
                                   'ScaleRatio': 0.00001,
                                   'ScaleOffset': 0},
                           'dst': {}}]

        self.GeolocVRT.create_bands(GeolocMetaDict)

        GeolocObject = Geolocation(x_vrt=self.GeolocVRT,
                                        y_vrt=self.GeolocVRT,
                                        # x = lon, y = lat
                                        x_band=1, y_band=2,
                                        line_offset=0, pixel_offset=0,
                                        line_step=1, pixel_step=1)

        # create empty VRT dataset with geolocation only
        # x_size, y_size, geo_transform, projection, gcps=None, gcp_projection='', **kwargs
        self._init_from_dataset_params(subDataset.RasterXSize, subDataset.RasterYSize,
                                        (0,1,0,subDataset.RasterYSize,0,-1),
                                        GeolocObject.d['SRS'])
        self._add_geolocation(GeolocObject)

        # Scale and NODATA should ideally be taken directly from raw file
        metaDict = [{'src': {'SourceFilename': ('NETCDF:"' + filename +
                                                '":wind_speed'),
                             'ScaleRatio': 0.01,
                             'NODATA': -32767},
                     'dst': {'name': 'windspeed',
                             'wkv': 'wind_speed'}
                     },
                    {'src': {'SourceFilename': ('NETCDF:"' + filename +
                                                '":wind_dir'),
                             'ScaleRatio': 0.1,
                             'NODATA': -32767},
                     'dst': {'name': 'winddirection',
                             'wkv': 'wind_from_direction'}}]

        self.create_bands(metaDict)

        # This should not be necessary
        # - should be provided by GeolocationArray!
        self.dataset.SetProjection(GeolocObject.d['SRS'])

        # Add time
        startTime = datetime.datetime(int(iFileName[6:10]),
                                      int(iFileName[10:12]),
                                      int(iFileName[12:14]),
                                      int(iFileName[15:17]),
                                      int(iFileName[17:19]),
                                      int(iFileName[19:21]))
        # Adding valid time to dataset
        self.dataset.SetMetadataItem('time_coverage_start', startTime.isoformat())
        self.dataset.SetMetadataItem('time_coverage_end', startTime.isoformat())

        # Get dictionary describing the instrument and platform according to
        # the GCMD keywords
        mm = pti.get_gcmd_instrument('ascat')
        ee = pti.get_gcmd_platform('metop-a')

        # TODO: Validate that the found instrument and platform are indeed what
        # we want....

        self.dataset.SetMetadataItem('instrument', json.dumps(mm))
        self.dataset.SetMetadataItem('platform', json.dumps(ee))



# Name:        mapper_pathfinder52
# Purpose:     Mapping for NOAA AVHRR PATHFINDER52 DATA
# Authors:      Anton Korosov, Dmitry Petrenko
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
from dateutil.parser import parse

import numpy as np

import nansat.vrt as vrt
from nansat.nsr import NSR
from nansat.exceptions import WrongMapperError


class Mapper(vrt.VRT):
    ''' Mapper PATHFINDER (local files)

    TODO:
    * remote files
    '''

    def __init__(self, filename, gdalDataset, gdalMetadata, minQual=4,
                 **kwargs):
        ''' Create VRT '''

        if not 'AVHRR_Pathfinder-PFV5.2' in filename:
            raise WrongMapperError(filename)

        subDatasets = gdalDataset.GetSubDatasets()
        metaDict = []
        sstName = ''

        for subDataset in subDatasets:
            subDatasetName = subDataset[0].split(':')[2]

            if '//' in subDatasetName:
                h5Style = True
            else:
                h5Style = False

            if h5Style:
                subDatasetName = subDatasetName.replace('//', '')

            if subDatasetName == 'quality_level':
                qualName = subDataset[0]

            subGDALDataset = vrt.gdal.Open(subDataset[0])
            subGDALMetadata = subGDALDataset.GetRasterBand(1).GetMetadata()
            if h5Style:
                metaPrefix = subDatasetName + '_'
            else:
                metaPrefix = ''

            subWKV = subGDALMetadata.get(metaPrefix + 'standard_name', '')
            subScaleRatio = subGDALMetadata.get(metaPrefix + 'scale_factor',
                                                '1')
            subScaleOffset = subGDALMetadata.get(metaPrefix + 'add_offset',
                                                 '0')
            metaEntry = {'src': {'SourceFilename': subDataset[0],
                                 'sourceBand': 1,
                                 'ScaleRatio': subScaleRatio,
                                 'ScaleOffset': subScaleOffset},
                         'dst': {'wkv': subWKV}}

            # append band metadata to metaDict
            metaDict.append(metaEntry)

        # create empty VRT dataset with geolocation only
        self._init_from_gdal_dataset(subGDALDataset)

        # add mask
        if qualName != '':
            qualDataset = vrt.gdal.Open(qualName)
            qualArray = qualDataset.ReadAsArray()
            qualArray[qualArray < minQual] = 1
            qualArray[qualArray >= minQual] = 128
            self.band_vrts = {'maskVRT': vrt.VRT(array=qualArray.astype('int8'))}
            metaDict.append({'src': {'SourceFilename': (self.
                                                        band_vrts['maskVRT'].
                                                        filename),
                                     'SourceBand': 1,
                                     'SourceType': 'SimpleSource',
                                     'DataType': 1},
                             'dst': {'name': 'mask'}})

        # add bands with metadata and corresponding values to the empty VRT
        self.create_bands(metaDict)

        # append fixed projection and geotransform
        self.dataset.SetProjection(NSR().wkt)
        self.dataset.SetGeoTransform((-180, 0.0417, 0, 90, 0, -0.0417))

        # set TIMEstart_time
        if h5Style:
            startTimeKey = 'start_time'
        else:
            startTimeKey = 'NC_GLOBAL#start_time'
        self.dataset.SetMetadataItem('time_coverage_start', subGDALDataset.GetMetadataItem(startTimeKey))



# Name:    vrt.py
# Purpose: Container of VRT classes
# Authors:      Asuka Yamakawa, Anton Korosov, Knut-Frode Dagestad,
#               Morten W. Hansen, Alexander Myasoyedov,
#               Dmitry Petrenko, Evgeny Morozov
# Created:      29.06.2011
# Copyright:    (c) NERSC 2011 - 2018
# Licence:
# This file is part of NANSAT.
# NANSAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
# http://www.gnu.org/licenses/gpl-3.0.html
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
from __future__ import absolute_import, unicode_literals, division
import os
import tempfile
from string import Template, ascii_uppercase, digits
from random import choice
import warnings
import pythesint as pti

import osr
import gdal
import numpy as np

from nansat.node import Node
from nansat.nsr import NSR
from nansat.geolocation import Geolocation
from nansat.tools import add_logger, numpy_to_gdal_type, gdal_type_to_offset, remove_keys

from nansat.exceptions import NansatProjectionError
from nansat.warnings import NansatFutureWarning

class VRT(object):
    """Wrapper around GDAL VRT-file

    The GDAL VRT-file is an XML-file. It contains all metadata, geo-reference
    information and information ABOUT each band including band metadata,
    reference to the bands in the source file.
    VRT-class performs all operation on VRT-files: create, copy, modify,
    read, write, add band, add GeoTransform, set Projection, etc. It uses
    either GDAL methods for these operations (e.g. Create, AddBand,
    SetMetadata, AutoCreateWarpedVRT, etc.) or reads/writes the XML-file
    directly (e.g. remove_geotransform, get_warped_vrt, etc).

    The core of the VRT object is GDAL dataset <self.dataset> generated
    by the GDAL VRT-Driver. The respective VRT-file is located in /vismem
    and has a random name.

    GDAL data model doesn't have place for geolocaion arrays therefore
    VRT-object has instance of Geolocation (self.geolocation)
    an object to keep information about Geolocation metadata:
    reference to file with source data, pixel and line step and offset, etc.

    Domain has an instance of VRT-class <self.vrt>. It keeps only geo-
    reference information.

    All Mappers inherit from VRT. When Nansat opens a file it loops through
    list of mappers, selects the one appropriate for the input file,
    and creates an instance of Mapper.

    Nansat has one instances of Mapper-class (>=VRT-class): self.vrt.
    It holds VRT-file in original projection (derived from the
    input file). After most of the operations with Nansat object
    (e.g. reproject, crop, resize, add_band) self.vrt is replaced with a new
    VRT object which has reference to the previous VRT object inside (self.vrt.vrt).

    Parameters
    ----------
    x_size : int
        width of self.dataset
    y_size : int
        arguments for VRT()
    metadata : dict
        dictionray with metadata keys (str) and values (str)
    nomem : bool
        don't create VRT in VSI memory?

    Notes
    --------
    adds self.logger, self.driver, self.filename, self.band_vrts, self.tps, self.vrt
    adds self.dataset - GDAL Dataset without bands and with size=(x_zie, y_size)
    adds metadata to self.dataset
    writes VRT file content to self.filename

    """
    INIT_GDALDATASET_WARNING = ('VRT(gdalDataset=...) will be disabled in Nansat 1.1. '
                                'Use VRT.from_gdal_dataset(...).')
    INIT_VRTDATASET_WARNING = ('VRT(vrtDataset=...) will be disabled in Nansat 1.1. '
                                'Use VRT.copy_dataset(...).')
    INIT_ARRAY_WARNING = ('VRT(array=...) will be disabled in Nansat 1.1. '
                          'Use VRT.from_array(...).')
    INIT_DATASETPARAMS_WARNING = ('VRT(srcGeoTransform, srcProjection, srcRasterXSize, '
                                  'srcRasterYSize, srcGCPs, srcGCPProjection) '
                                  'will be disabled in Nansat 1.1. '
                                  'Use VRT.from_dataset_params(...).')
    INIT_GEOLOCATIONARRAY_WARNING = ('VRT(geolocationArray=...) will be disabled in Nansat 1.1. '
                                     'Use vrt._add_geolocation(...).')
    INIT_LATLON_WARNING = ('VRT(lat=..., lon=...) will be disabled in Nansat 1.1. '
                           'Use VRT.from_lonlat(...).')
    INIT_SRCMETADATA_WARNING = ('VRT(srcMetadata=...) will be disabled in Nansat 1.1. '
                                'Use VRT(metadata=...).')
    COMPLEX_SOURCE_XML = Template('''
            <$SourceType>
                <SourceFilename relativeToVRT="0">$Dataset</SourceFilename>
                <SourceBand>$SourceBand</SourceBand>
                <NODATA>$NODATA</NODATA>
                <ScaleOffset>$ScaleOffset</ScaleOffset>
                <ScaleRatio>$ScaleRatio</ScaleRatio>
                <LUT>$LUT</LUT>
                <SrcRect xOff="$xOff" yOff="$yOff" xSize="$xSize" ySize="$ySize"/>
                <DstRect xOff="0" yOff="0" xSize="$xSize" ySize="$ySize"/>
            </$SourceType> ''')

    RAW_RASTER_BAND_SOURCE_XML = Template('''
            <VRTDataset rasterXSize="$XSize" rasterYSize="$YSize">
              <VRTRasterBand dataType="$DataType"
                band="$BandNum" subClass="VRTRawRasterBand">
                <SourceFilename relativeToVRT="0">$SrcFileName</SourceFilename>
                <ImageOffset>0</ImageOffset>
                <PixelOffset>$PixelOffset</PixelOffset>
                <LineOffset>$LineOffset</LineOffset>
              </VRTRasterBand>
            </VRTDataset> ''')

    REPROJECT_TRANSFORMER = Template('''
        <ReprojectTransformer>
          <ReprojectionTransformer>
            <SourceSRS>$SourceSRS</SourceSRS>
            <TargetSRS>$TargetSRS</TargetSRS>
          </ReprojectionTransformer>
        </ReprojectTransformer> ''')

    # instance attributes
    filename = ''
    vrt = None
    dataset = None
    logger = None
    driver = None
    band_vrts = None
    tps = None

    @classmethod
    def from_gdal_dataset(cls, gdal_dataset, **kwargs):
        """Create VRT from GDAL Dataset with the same size/georeference but wihout bands.

        Parameters
        ----------
        gdal_dataset : gdal.Dataset
            input GDAL dataset
        **kwargs : dict
            arguments for VRT()

        Returns
        -------
        vrt : VRT

        """
        vrt = cls.__new__(cls)
        vrt._init_from_gdal_dataset(gdal_dataset, **kwargs)
        return vrt

    @classmethod
    def from_dataset_params(cls, x_size, y_size, geo_transform, projection,
                            gcps, gcp_projection, **kwargs):
        """Create VRT from GDAL Dataset parameters

        Create VRT with dataset wihout bands but with size/georeference corresponding to
        input parameters.

        Parameters
        ----------
        x_size : int
            X-size of dataset
        y_size : int
            Y-size of dataset
        geotransform : tuple with 6 floats
            informaton on affine transformtaion
        projection : str
            WKT representation of spatial reference system
        gcps : tuple or list with GDAL GCP objects
            GDAL Ground Control Points
        gcp_projection : str
            WKT representation of GCPs spatial reference system
        **kwargs : dict
            arguments for VRT()

        Returns
        -------
        vrt : VRT

        """
        vrt = cls.__new__(cls)
        vrt._init_from_dataset_params(x_size, y_size, geo_transform, projection,
                         gcps, gcp_projection, **kwargs)
        return vrt

    @classmethod
    def from_array(cls, array, **kwargs):
        """Create VRT from numpy array with dataset wih one band but without georeference.

        Parameters
        ----------
        array : numpy.ndarray
            array with data
        **kwargs : dict
            arguments for VRT()

        Returns
        -------
        vrt : VRT

        """
        vrt = cls.__new__(cls)
        vrt._init_from_array(array, **kwargs)
        return vrt

    @classmethod
    def from_lonlat(cls, lon, lat, **kwargs):
        """Create VRT from longitude, latitude arrays

        Create VRT with dataset without bands but with GEOLOCATION metadata and Geolocation
        object. Geolocation contains 2 2D arrays with lon/lat values given at regular pixel/line
        steps.

        Parameters
        ----------
        lon : numpy.ndarray
            array with longitudes
        lat : numpy.ndarray
            array with latitudes
        **kwargs : dict
            arguments for VRT()

        Returns
        -------
        vrt : VRT

        """
        vrt = cls.__new__(cls)
        vrt._init_from_lonlat(lon, lat)
        return vrt

    @classmethod
    def copy_dataset(cls, gdal_dataset, **kwargs):
        """Create VRT with bands and georefernce as a full copy of input GDAL Dataset

        Parameters
        ----------
        gdal_dataset : GDAL.Dataset
            input dataset
        **kwargs : dict
            arguments for VRT()

        Returns
        -------
        vrt : VRT

        """
        vrt = cls.__new__(cls)
        vrt._copy_from_dataset(gdal_dataset, **kwargs)
        return vrt

    def __init__(self, x_size=1, y_size=1, metadata=None, nomem=False, **kwargs):
        """Init VRT object with all attributes"""
        if isinstance(x_size, gdal.Dataset):
            kwargs['gdalDataset'] = x_size
        if metadata is None:
            metadata = dict()
        # init from obsolete params
        if self._init_from_old_params(metadata=metadata, nomem=nomem, **kwargs):
            return

        # essential attributes
        self.logger = add_logger('Nansat')
        self.driver = gdal.GetDriverByName(str('VRT'))
        self.filename = str(VRT._make_filename(nomem=nomem))
        self.band_vrts = dict()
        self.tps = False
        self.vrt = None

        # create dataset
        self.dataset = self.driver.Create(self.filename, x_size, y_size, bands=0)
        self.dataset.SetMetadata(metadata)
        self.dataset.FlushCache()

    def _init_from_gdal_dataset(self, gdal_dataset, **kwargs):
        """Init VRT from GDAL Dataset with the same size/georeference but wihout bands/metadata.

        Parameters
        ----------
        gdal_dataset : gdal.Dataset
            input GDAL dataset
        **kwargs : dict
            arguments for VRT()

        Notes
        -------
        self - adds all VRT attributes
        self.dataset - sets size and georeference

        """
        # WORKAROUND for not providing metadata explicitly. Should be avoided.
        # get metadata from input and gdal_dataset
        # metadata = kwargs.pop('metadata', dict())
        # metadata.update(gdal_dataset.GetMetadata())
        # set dataset parameters and metadata
        VRT.__init__(self, gdal_dataset.RasterXSize, gdal_dataset.RasterYSize, **kwargs)
        self.dataset.SetGCPs(gdal_dataset.GetGCPs(), gdal_dataset.GetGCPProjection())
        self.dataset.SetProjection(gdal_dataset.GetProjection())
        self.dataset.SetGeoTransform(gdal_dataset.GetGeoTransform())
        self._add_geolocation(Geolocation.from_dataset(gdal_dataset))
        self.dataset.SetMetadataItem(str('filename'), self.filename)

        # write XML file contents
        self.dataset.FlushCache()

    def _init_from_dataset_params(self, x_size, y_size, geo_transform, projection,
                                    gcps=None, gcp_projection='', **kwargs):
        """Init VRT from GDAL Dataset parameters

        Init VRT with dataset wihout bands but with size/georeference corresponding to
        input parameters.

        Parameters
        ----------
        x_size : int
            X-size of dataset
        y_size : int
            Y-size of dataset
        geotransform : tuple with 6 floats
            informaton on affine transformtaion
        projection : str
            WKT representation of spatial reference system
        gcps : tuple or list with GDAL GCP objects
            GDAL Ground Control Points
        gcp_projection : str
            WKT representation of GCPs spatial reference system
        **kwargs : dict
            arguments for VRT()

        Notes
        -------
        self - adds all VRT attributes
        self.dataset - sets size and georeference

        """
        # x_size, y_size, geo_transform, projection, gcps=None, gcp_projection='', **kwargs
        VRT.__init__(self, x_size, y_size, **kwargs)
        # set dataset (geo-)metadata
        self.dataset.SetProjection(str(projection))
        self.dataset.SetGeoTransform(geo_transform)
        if isinstance(gcps, (list, tuple)):
            self.dataset.SetGCPs(gcps, str(gcp_projection))
        self.dataset.SetMetadataItem(str('filename'), self.filename)

        # write file contents
        self.dataset.FlushCache()

    def _init_from_array(self, array, **kwargs):
        """Init VRT from numpy array with dataset wih one band but without georeference.

        Write contents of the array into flat binary file (VSI)
        Write VRT file with RawRastesrBand, which points to the binary file
        Open the VRT file as self.dataset with GDAL

        Parameters
        ----------
        array : numpy.ndarray
            array with data
        **kwargs : dict
            arguments for VRT()

        Notes
        ---------
        binary file is written (VSI)
        VRT file is written (VSI)
        self - adds all VRT attributes
        self.dataset is updated

        """
        VRT.__init__(self, **kwargs)
        # create flat binary file (in VSI) from numpy array
        array_type = array.dtype.name
        array_shape = array.shape
        binary_file = self.filename.replace('.vrt', '.raw')
        ofile = gdal.VSIFOpenL(str(binary_file), str('wb'))
        gdal.VSIFWriteL(array.tostring(), len(array.tostring()), 1, ofile)
        gdal.VSIFCloseL(ofile)
        array = None

        # convert Numpy datatype to gdal datatype and pixel offset
        gdal_data_type = numpy_to_gdal_type[array_type]
        pixel_offset = gdal_type_to_offset[gdal_data_type]

        # create XML contents of VRT-file
        line_offset = str(int(pixel_offset) * array_shape[1])
        contents = self.RAW_RASTER_BAND_SOURCE_XML.substitute(
            XSize=array_shape[1],
            YSize=array_shape[0],
            DataType=gdal_data_type,
            BandNum=1,
            SrcFileName=binary_file,
            PixelOffset=pixel_offset,
            LineOffset=line_offset)

        # write XML contents to VRT-file
        self.write_xml(contents)
        self.dataset.SetMetadataItem(str('filename'), self.filename)
        self.dataset.FlushCache()

    def _init_from_lonlat(self, lon, lat, **kwargs):
        """Init VRT from longitude, latitude arrays

        Init VRT with dataset without bands but with GEOLOCATION metadata and Geolocation
        object. Geolocation contains 2 2D arrays with lon/lat values given at regular pixel/line
        steps.

        Parameters
        ----------
        lon : numpy.ndarray
            array with longitudes
        lat : numpy.ndarray
            array with latitudes
        **kwargs : dict
            arguments for VRT() and VRT._lonlat2gcps

        Notes
        -------
        self - adds all VRT attributes
        self.dataset - sets size and georeference
        self.geolocation - add Geolocation object with all attributes

        """
        VRT.__init__(self, lon.shape[1], lon.shape[0], **kwargs)
        self.dataset.SetGCPs(VRT._lonlat2gcps(lon, lat, **kwargs), NSR().wkt)
        self._add_geolocation(Geolocation(VRT.from_array(lon), VRT.from_array(lat)))
        self.dataset.SetMetadataItem(str('filename'), self.filename)
        self.dataset.FlushCache()

    def _copy_from_dataset(self, gdal_dataset, **kwargs):
        """Init VRT with bands and georefernce as a full copy of input GDAL Dataset

        Parameters
        ----------
        gdal_dataset : GDAL.Dataset
            input dataset
        **kwargs : dict
            arguments for VRT()

        Notes
        --------
        self - adds all VRT attributes
        self.dataset - sets size and georeference

        """
        # set dataset geo-metadata
        VRT.__init__(self, gdal_dataset.RasterXSize, gdal_dataset.RasterYSize, **kwargs)
        self.dataset = self.driver.CreateCopy(self.filename, gdal_dataset)
        self.dataset.SetMetadataItem(str('filename'), self.filename)

        # write XMl file contents
        self.dataset.FlushCache()

    def __del__(self):
        """Destructor deletes VRT and RAW files"""
        self.dataset = None
        gdal.Unlink(self.filename)
        gdal.Unlink(self.filename.replace('vrt', 'raw'))

    def __repr__(self):
        str_out = os.path.split(self.filename)[1]
        if self.vrt is not None:
            str_out += '=>%s' % self.vrt.__repr__()
        return str_out

    def _create_complex_bands(self, filenames):
        """Create bands with complex data type bands with real and imag components
        Parameters
        ---------
            filenames : list of used filenames

        Notes
        --------
        self.dataset - adds complex bands; removes real and imag bands

        """
        rm_bands = []
        # loop to find real data band
        for i in range(self.dataset.RasterCount):
            band = self.dataset.GetRasterBand(i + 1)
            band_name = band.GetMetadataItem(str('name'))
            if band_name.endswith('_real'):
                real_band_no = i
                real_band_type = band.GetMetadataItem(str('DataType'))
                complex_band_name = band_name.replace('_real', '')
                # loop to find imag data band
                for j in range(self.dataset.RasterCount):
                    band = self.dataset.GetRasterBand(j + 1)
                    band_name = band.GetMetadataItem(str('name'))
                    # find an imaginary data band corresponding to the real
                    # data band and create complex data band from the bands
                    if band_name == complex_band_name + '_imag':
                        imag_band_no = j
                        imag_band_type = band.GetMetadataItem(str('DataType'))
                        dst = band.GetMetadata()
                        dst['name'] = complex_band_name
                        dst['PixelFunctionType'] = 'ComplexData'
                        dst['dataType'] = 10
                        src = [{'SourceFilename': filenames[real_band_no],
                                'SourceBand':  1,
                                'DataType': real_band_type},
                               {'SourceFilename': filenames[imag_band_no],
                                'SourceBand': 1,
                                'DataType': imag_band_type}]
                        self.create_band(src, dst)
                        self.dataset.FlushCache()
                        rm_bands.append(real_band_no + 1)
                        rm_bands.append(imag_band_no + 1)
                        break

        # Delete real and imaginary bands
        if len(rm_bands) != 0:
            self.delete_bands(rm_bands)

    def _add_swath_mask_band(self):
        """Create a new band where all values = 1

        Notes
        ---------
        Single band 'swathmask' with ones is added to the self.dataset

        """
        self.create_band(
            src=[{
                'SourceFilename': self.filename,
                'SourceBand':  1,
                'DataType': gdal.GDT_Byte}],
            dst={
                'dataType': gdal.GDT_Byte,
                'wkv': 'swath_binary_mask',
                'PixelFunctionType': 'OnesPixelFunc',
            })

    def _add_geolocation(self, geolocation):
        """ Add GEOLOCATION to the VRT

        Parameters
        -----------
            geolocation: Geolocation
                with grids of X/Y coordinates

        Notes
        --------
        Add geolocation to self
        Sets GEOLOCATION metadata

        """
        self.geolocation = geolocation
        self.dataset.SetMetadata(geolocation.data, str('GEOLOCATION'))
        self.dataset.FlushCache()

    def _remove_geolocation(self):
        """ Remove GEOLOCATION from the VRT

        Notes
        --------
        Set self.geolocationArray to None
        Sets GEOLOCATION metadata to ''

        """
        del self.geolocation

        # add GEOLOCATION metadata (empty if geolocation is empty)
        self.dataset.SetMetadata(str(''), str('GEOLOCATION'))
        self.dataset.FlushCache()

    def _remove_geotransform(self):
        """Remove GeoTransfomr from VRT Object

        Notes
        ---------
        The tag <GeoTransform> is revoved from the VRT-file

        """
        # read XML content from VRT
        # find and remove GeoTransform
        node0 = Node.create(self.xml)
        node0.delNode('GeoTransform')
        # Write the modified elemements back into temporary VRT
        self.write_xml(node0.rawxml())

    def _set_fake_gcps(self, dst_srs, dst_gcps, skip_gcps):
        """Create GCPs with reference self.pixel/line ==> dst.pixel/line and set to self.dataset

        GCPs from a destination image (dst_gcps) are converted to a gcp of source
        image (src_gcps) this way:

        src_gcpPixel = srcPixel
        src_gcpLine = srcLine
        src_gcpX = dstGCPPixel = f(src_srs, dstGCPX, dstGCPY)
        src_gcpY = dstGCPLine = f(src_srs, dstGCPX, dstGCPY)

        Parameters
        -----------
        dst_srs : str
            destination SRS
        dst_gcps : list or tuple
            GDAL GCPs
        skip_gcps : int
            number of GCPs to skip

        Returns
        --------
        srs : str or None
            if GCPs are set, returns None
            if GCPs are absent, retuns dst_srs

        """
        if len(dst_gcps) == 0:
            return dst_srs
        # create transformer. converts lat/lon to pixel/line of SRC image
        src_transformer = gdal.Transformer(self.dataset, None,
                                          ['SRC_SRS=' + self.get_projection(),
                                           'DST_SRS=' + NSR(dst_srs).wkt])

        # create 'fake' GCPs
        fake_gcps = []
        for g in dst_gcps[::skip_gcps]:
            # transform DST lat/lon to SRC pixel/line
            succ, point = src_transformer.TransformPoint(1, g.GCPX, g.GCPY)
            srcPixel = point[0]
            srcLine = point[1]

            # swap coordinates in GCPs:
            # pix1/line1 -> lat/lon  =>=>  pix2/line2 -> pix1/line1
            fake_gcps.append(gdal.GCP(g.GCPPixel, g.GCPLine,
                                     0, srcPixel, srcLine))

        self.dataset.SetGCPs(fake_gcps, NSR('+proj=stere').wkt)
        return None

    def _set_geotransform_for_resize(self):
        """Prepare VRT.dataset for resizing."""
        self.dataset.SetMetadata(str(''), str('GEOLOCATION'))
        self.dataset.SetGCPs([], str(''))
        self.dataset.SetGeoTransform((0, 1, 0, self.dataset.RasterYSize, 0, -1))

    def _set_gcps_geolocation_geotransform(self):
        """Prepare VRT.dataset for warping."""
        # Select if GEOLOCATION, or GCPs, or GeoTransform from the original dataset are used
        if hasattr(self, 'geolocation') and len(self.geolocation.data) > 0:
            # use GEOLOCATION by default (remove GCP and GeoTransform)
            self.dataset.SetGCPs([], str(''))
            self._remove_geotransform()
        elif len(self.dataset.GetGCPs()) > 0:
            # otherwise fallback to GCPs (remove Geolocation and GeoTransform)
            self.dataset.SetMetadata(str(''), str('GEOLOCATION'))
            self._remove_geotransform()
        else:
            # otherwise fallback to GeoTransform in input VRT (remove Geolocation and GCP)
            self.dataset.SetMetadata(str(''), str('GEOLOCATION'))
            self.dataset.SetGCPs([], str(''))
        self.dataset.FlushCache()

    def _update_warped_vrt_xml(self, x_size, y_size, geo_transform, block_size, working_data_type):
        """Update rasterXsize, rasterYsize, geotransform, block_size and working_data_type"""
        node0 = Node.create(str(self.xml))
        node0.replaceAttribute('rasterXSize', str(x_size))
        node0.replaceAttribute('rasterYSize', str(y_size))

        if geo_transform is not None:
            invGeotransform = gdal.InvGeoTransform(geo_transform)
            # convert proper string style and set to the GeoTransform element
            node0.node('GeoTransform').value = str(geo_transform).strip('()')
            node0.node('DstGeoTransform').value = str(geo_transform).strip('()')
            node0.node('DstInvGeoTransform').value = (
                str(invGeotransform[1]).strip('()'))

            if node0.node('SrcGeoLocTransformer'):
                node0.node('BlockXSize').value = str(x_size)
                node0.node('BlockYSize').value = str(y_size)

            if block_size is not None:
                node0.node('BlockXSize').value = str(block_size)
                node0.node('BlockYSize').value = str(block_size)

            if working_data_type is not None:
                node0.node('WorkingDataType').value = working_data_type

        # overwrite XML file with updated size, geotranform, etc
        self.write_xml(node0.rawxml())

    def _create_band_name(self, dst):
        """Create band name based on destination band dictionary <dst>"""
        band_name = dst.get('name', None)

        # try to get metadata from WKV using PyThesInt if it exists
        wkv_dst = dst.get('wkv', None)
        try:
            wkv_pti = pti.get_wkv_variable(str(wkv_dst))
        except IndexError:
            # IndexError is raised when PyThesInt doesn't find the requested WKV.
            # In that case and empty dict without any metadata is created
            wkv = {}
        else:
            # If WKV was found by PyThesInt, a dict with metadata is created
            wkv = dict(wkv_pti)

        if band_name is None:
            band_name = wkv.get('short_name', 'band')
            if 'suffix' in dst:
                 band_name += '_' + dst['suffix']

        # create list of available bands (to prevent duplicate names)
        band_names = [self.dataset.GetRasterBand(i + 1).GetMetadataItem(str('name'))
                        for i in range(self.dataset.RasterCount)]

        # check if name already exist and add '_NNNN'
        dst_band_name = band_name
        for n in range(9999):
            if dst_band_name not in band_names:
                break
            dst_band_name = '%s_%04d' % (band_name, n)

        return dst_band_name, wkv

    def _find_complex_band(self):
        """Find complex data bands"""
        # find complex bands
        for i in range(1, self.dataset.RasterCount+1):
            if self.dataset.GetRasterBand(i).DataType in [8, 9, 10, 11]:
                return i
        return None

    def _init_from_old_params(self, **kwargs):
        """Check obsolete input parameters for backward compatibility"""
        old_param_names = ['gdalDataset',
                           'vrtDataset',
                           'array',
                           'srcGeoTransform',
                           'srcProjection',
                           'srcRasterXSize',
                           'srcRasterYSize',
                           'srcGCPs',
                           'srcGCPProjection',
                           'lon',
                           'lat']

        old_params_used = False
        # run through input params and select the matching one
        for arg in old_param_names:
            if arg in kwargs:
                old_params_used = True
                break
        if not old_params_used:
            return False

        # parameters for self._init_from_dataset_params()
        dataset_params = (kwargs.get('srcRasterXSize', None),
                          kwargs.get('srcRasterYSize', None),
                          kwargs.get('srcGeoTransform', (0, 1, 0, 0, 0, -1)),
                          kwargs.get('srcProjection', ''),
                          kwargs.get('srcGCPs', list()),
                          kwargs.get('srcGCPProjection', ''))

        # conversion from old input parameters to warning, new function and new parameters
        new_warnings_and_funcs = [
        [self.INIT_GDALDATASET_WARNING, self._init_from_gdal_dataset, (kwargs.get('gdalDataset', None),)],
        [self.INIT_VRTDATASET_WARNING, self._copy_from_dataset, (kwargs.get('vrtDataset', None),)],
        [self.INIT_ARRAY_WARNING, self._init_from_array, (kwargs.get('array'),)],
        [self.INIT_DATASETPARAMS_WARNING, self._init_from_dataset_params, dataset_params],
        [self.INIT_DATASETPARAMS_WARNING, self._init_from_dataset_params, dataset_params],
        [self.INIT_DATASETPARAMS_WARNING, self._init_from_dataset_params, dataset_params],
        [self.INIT_DATASETPARAMS_WARNING, self._init_from_dataset_params, dataset_params],
        [self.INIT_DATASETPARAMS_WARNING, self._init_from_dataset_params, dataset_params],
        [self.INIT_DATASETPARAMS_WARNING, self._init_from_dataset_params, dataset_params],
        [self.INIT_LATLON_WARNING, self._init_from_lonlat, (kwargs.get('lon', None), kwargs.get('lat', None))],
        [self.INIT_LATLON_WARNING, self._init_from_lonlat, (kwargs.get('lon', None), kwargs.get('lat', None))],
        ]

        old2new = dict(zip(old_param_names, new_warnings_and_funcs))

        # get metadata
        metadata = kwargs['metadata']
        if 'srcMetadata' in kwargs:
            warnings.warn(self.INIT_SRCMETADATA_WARNING, NansatFutureWarning)
            metadata.update(kwargs['srcMetadata'])

        # get metadata also from input gdalDataset
        if arg == 'gdalDataset':
            metadata.update(kwargs['gdalDataset'].GetMetadata())

        # raise warning
        warnings.warn(old2new[arg][0], NansatFutureWarning)
        # call function
        old2new[arg][1](*old2new[arg][2], metadata=metadata, nomem=kwargs[str('nomem')])

        return old_params_used

    def leave_few_bands(self, bands=None):
        """Leave only given bands in VRT"""
        if bands is None:
            return

        rm_bands = []
        for i in range(1, self.dataset.RasterCount+1):
            band_name = self.dataset.GetRasterBand(i).GetMetadata().get('name', '')
            if i not in bands and band_name not in bands:
                rm_bands.append(i)
        # delete bands from VRT
        self.delete_bands(rm_bands)

    def split_complex_bands(self):
        """Recursevly find complex bands and relace by real and imag components"""
        rm_metadata= ['dataType', 'PixelFunctionType']

        i = self._find_complex_band()
        if i is None:
            return

        band = self.dataset.GetRasterBand(i)
        band_array = band.ReadAsArray()
        band_metadata_orig = remove_keys(band.GetMetadata(), rm_metadata)
        band_name_orig = band_metadata_orig.get('name', 'complex_%003d'%i)
        # Copy metadata, modify 'name' and create VRTs for real and imag parts of each band
        band_description = []
        for part in ['real', 'imag']:
            band_metadata = band_metadata_orig.copy()
            band_name = band_name_orig + '_' + part
            band_metadata['name'] = band_name
            self.band_vrts[band_name] = VRT.from_array(eval('band_array.' + part))
            band_description.append({'src': {
                     'SourceFilename': self.band_vrts[band_name].filename,
                     'SourceBand':  1},
                     'dst': band_metadata})
        # create bands with parts
        self.create_bands(band_description)
        # delete the complex band
        self.delete_band(i)
        # find and split more complex bands
        self.split_complex_bands()

    def create_geolocation_bands(self):
        """Create bands from Geolocation"""
        if hasattr(self, 'geolocation') and len(self.geolocation.data) > 0:
            self.create_band(
                {'SourceFilename': self.geolocation.data['X_DATASET'],
                 'SourceBand': int(self.geolocation.data['X_BAND'])},
                {'wkv': 'longitude',
                 'name': 'longitude'})
            self.create_band(
                {'SourceFilename': self.geolocation.data['Y_DATASET'],
                 'SourceBand': int(self.geolocation.data['Y_BAND'])},
                {'wkv': 'latitude',
                 'name': 'latitude'})
        self.dataset.FlushCache()

    def fix_band_metadata(self, rm_metadata):
        """Add NETCDF_VARNAME and remove <rm_metadata> in metadata for each band"""
        for iBand in range(self.dataset.RasterCount):
            band = self.dataset.GetRasterBand(iBand + 1)
            metadata = remove_keys(band.GetMetadata(), rm_metadata)
            if 'name' in metadata:
                metadata['NETCDF_VARNAME'] = metadata['name']
            band.SetMetadata(metadata)
        self.dataset.FlushCache()

    def fix_global_metadata(self, rm_metadata):
        """Remove unwanted global metadata and escape special characters"""
        metadata = remove_keys(self.dataset.GetMetadata(), rm_metadata)
        # Apply escaping to metadata strings to preserve special characters (in XML/HTML format)
        metadata_escaped = {}
        for key, val in list(metadata.items()):
            # Keys not escaped - this may be changed if needed...
            metadata_escaped[key] = gdal.EscapeString(val, gdal.CPLES_XML)
        self.dataset.SetMetadata(metadata_escaped)
        self.dataset.FlushCache()

    def hardcopy_bands(self):
        """Make 'hardcopy' of bands: evaluate array from band and put into original band"""
        bands = range(1, self.dataset.RasterCount+1)
        for i in bands:
            self.band_vrts[i] = VRT.from_array(self.dataset.GetRasterBand(i).ReadAsArray())

        node0 = Node.create(str(self.xml))
        for i, iNode1 in enumerate(node0.nodeList('VRTRasterBand')):
            iNode1.node('SourceFilename').value = self.band_vrts[i+1].filename
            iNode1.node('SourceBand').value = str(1)
        self.write_xml(node0.rawxml())

    def prepare_export_gtiff(self, options):
        """Prepare dataset for export using GTiff driver"""
        if len(self.dataset.GetGCPs()) > 0:
            self._remove_geotransform()
        return options, False

    def prepare_export_netcdf(self, options, bottomup):
        """Prepare dataset for export using netCDF driver"""
        # set bottomup option
        if bottomup:
            options += ['WRITE_BOTTOMUP=NO']
        else:
            options += ['WRITE_BOTTOMUP=YES']
        gcps = self.dataset.GetGCPs()
        srs = str(self.get_projection())
        if len(gcps) > 0:
            self._remove_geotransform()
            self.dataset.SetMetadataItem(str('NANSAT_GCPProjection'),
                                         str(srs.replace(',', '|').replace('"', '&')))
            add_gcps = True
        else:
            self.dataset.SetMetadataItem(str('NANSAT_Projection'),
                                         str(srs.replace(',', '|').replace('"', '&')))

            # add GeoTransform metadata
            geo_transform_str = str(self.dataset.GetGeoTransform()).replace(',', '|')
            self.dataset.SetMetadataItem(str('NANSAT_GeoTransform'), str(geo_transform_str))
            add_gcps = False
        return options, add_gcps

    def copy(self):
        """Create and return a full copy of a VRT instance with new filenames

        If self.dataset has no bands, the copy is created also without bands.
        If self.dataset has bands, the copy is created from the dataset with all bands.
        If self has attribute 'vrt' (a sub-VRT object, result of get_super_vrt) it's contents is
        also copied into sub-VRT of the copy.
        Other attributes of self, such as tps flag and band_vrts are also copied.

        """
        if self.dataset.RasterCount == 0:
            new_vrt = VRT.from_gdal_dataset(self.dataset, metadata=self.dataset.GetMetadata())
        else:
            new_vrt = VRT.copy_dataset(self.dataset, metadata=self.dataset.GetMetadata())
            replace_filenames = [(self.filename, new_vrt.filename)]
            if self.vrt is not None:
                # recursive copy of sub-VRT object
                new_vrt.vrt = self.vrt.copy()
                replace_filenames.append((self.vrt.filename, new_vrt.vrt.filename))

            # change reference from original filenames to the new ones
            for replace_filename in replace_filenames:
                new_vrt_xml = new_vrt.xml.replace(os.path.basename(replace_filename[0]),
                                          os.path.basename(replace_filename[1]))
                new_vrt.write_xml(new_vrt_xml)

            # copy VRTs of bands
            new_vrt.band_vrts = dict(self.band_vrts)
        # copy the thin spline transformation option
        new_vrt.tps = bool(self.tps)

        return new_vrt

    @property
    def fileName(self):
        warnings.warn('VRT.fileName will be disabled in Nansat 1.1. Use VRT.filename instead.',
                NansatFutureWarning)
        return self.filename

    @property
    def xml(self):
        """Read XML content of the VRT-file using VSI

        Returns
        --------
        string : XMl Content which is read from the VSI file
        """
        self.dataset.FlushCache()
        return VRT.read_vsi(self.filename)

    def _create_bands(self, metadata_dict):
        warnings.warn('Method VRT._create_bands() will be disabled in Nansat 1.1. ' \
                'Use VRT.create_bands() instead.', NansatFutureWarning)
        self.create_bands(metadata_dict)

    def create_bands(self, metadata_dict):
        """ Generic function called from the mappers to create bands
        in the VRT dataset from an input dictionary of metadata

        Parameters
        ----------
        metadata_dict : list of dict with params of input bands and generated bands.
            Each dict has:
                'src' : dictionary with parameters of the sources:
                'dst' : dictionary with parameters of the generated bands

        Notes
        ---------
        Adds bands to the self.dataset based on info in metaDict

        See Also
        ---------
        VRT.create_band()

        """
        for band_dict in metadata_dict:
            src = band_dict['src']
            dst = band_dict.get('dst', None)
            self.create_band(src, dst)
            self.logger.debug('Creating band - OK!')
        self.dataset.FlushCache()

    def create_band(self, src, dst=None):
        """ Add band to self.dataset:

        Get parameters of the source band(s) from input
        Generate source XML for the VRT, add options of creating
        Call GDALDataset.AddBand
        Set source and options
        Add metadata

        Parameters
        ----------
        src : dict with parameters of sources:
            SourceFilename
            SourceBand
            ScaleRatio
            ScaleOffset
            NODATA
            LUT
            SourceType
            DataType
            ImageOffset (RawVRT)
            PixelOffset (RawVRT)
            LineOffset (RawVRT)
            ByteOrder (RawVRT)
            xSize
            ySize
        dst : dict with parameters of the created band
            name
            dataType
            wkv
            suffix
            AnyOtherMetadata
            PixelFunctionType: - band will be a pixel function defined by the
                                 corresponding name/value.
                                 In this case src may be list of
                                 dicts with parameters for each source.
                               - in case the dst band has a different datatype
                                 than the source band it is important to add a
                                 SourceTransferType parameter in dst
            SourceTransferType

        Returns
        --------
        name : string, name of the added band

        Examples
        --------
        vrt.create_band({'SourceFilename': filename, 'SourceBand': 1})
        vrt.create_band({'SourceFilename': filename, 'SourceBand': 2,
                          'ScaleRatio': 0.0001},
                         {'name': 'LAT', 'wkv': 'latitude'})
        vrt.create_band({'SourceFilename': filename, 'SourceBand': 2},
                         {'suffix': '670',
                          'wkv': 'brightness_temperature'})
        vrt.create_band([{'SourceFilename': filename, 'SourceBand': 1},
                          {'SourceFilename': filename, 'SourceBand': 1}],
                         {'PixelFunctionType': 'NameOfPixelFunction'})

        """
        self.logger.debug('INPUTS: %s, %s " ' % (str(src), str(dst)))
        # Make sure src is list, ready for loop
        if type(src) == dict:
            srcs = [src]
        elif type(src) in [list, tuple]:
            srcs = src
        else:
            raise ValueError('Wrong src type (%s)! Should be dict or list/tuple of dict'%type(src))

        # Check if dst is given, or create empty dict
        if dst is None:
            dst = {}

        srcs = list(map(VRT._make_source_bands_xml, srcs))
        options = VRT._set_add_band_options(srcs, dst)
        dst['dataType'] = VRT._get_dst_band_data_type(srcs, dst)
        dst['name'], wkv = self._create_band_name(dst)

        # Add Band
        self.dataset.AddBand(int(dst['dataType']), options=options)
        dst_raster_band = self.dataset.GetRasterBand(self.dataset.RasterCount)

        # Append sources to destination dataset
        if len(srcs) == 1 and srcs[0]['SourceBand'] > 0:
            # only one source
            dst_raster_band.SetMetadataItem(str('source_0'), str(srcs[0]['XML']),
                                            str('new_vrt_sources'))
        elif len(srcs) > 1:
            # several sources for PixelFunction
            metadataSRC = {}
            for i, src in enumerate(srcs):
                metadataSRC['source_%d' % i] = src['XML']
            dst_raster_band.SetMetadata(metadataSRC, str('vrt_sources'))

        # set metadata from WKV
        dst_raster_band = VRT._put_metadata(dst_raster_band, wkv)

        # set metadata from provided parameters
        # remove and add params
        dst['SourceFilename'] = srcs[0]['SourceFilename']
        dst['SourceBand'] = str(srcs[0]['SourceBand'])
        dst_raster_band = VRT._put_metadata(dst_raster_band, dst)

        # return name of the created band
        return dst['name']

    def write_xml(self, vsi_file_content=None):
        """Write XML content into a VRT dataset

        Parameters
        -----------
        vsi_fileContent: string, optional
            XML Content of the VSI file to write

        Notes
        ---------
        self.dataset
            If XML content was written, self.dataset is re-opened

        """
        vsi_file = gdal.VSIFOpenL(self.filename, str('w'))
        gdal.VSIFWriteL(str(vsi_file_content), len(vsi_file_content), 1, vsi_file)
        gdal.VSIFCloseL(vsi_file)
        # re-open self.dataset with new content
        self.dataset = gdal.Open(self.filename)

    def export(self, filename):
        """Export VRT file as XML into given <filename>"""
        self.driver.CreateCopy(filename, self.dataset)

    def _get_sub_filenames(self, gdal_dataset):
        """ Get filenames of subdatasets

        Parameters
        ----------
        gdal_dataset : gdal.Dataset
            Main gdal dataset containing sub-datasets

        Returns
        -------
        filenames : list
           List of filenames of each sub-dataset within the netCDF file

        """
        return [f[0] for f in gdal_dataset.GetSubDatasets()]

    def get_warped_vrt(self, dst_srs, x_size, y_size, geo_transform,
                       resample_alg=0,
                       dst_gcps=[],
                       skip_gcps=1,
                       block_size=None,
                       working_data_type=None,
                       resize_only=False):

        """Create warped (reprojected) VRT object

        1. Create a simple warped dataset using GDAL.AutoCreateWarpedVRT.
        2. Create VRT from the warped dataset.
        3. Modify the warped VRT according to the input options (size, geotransform, GCPs, etc)
        4. Keep the original VRT in the attribute vrt

        For reprojecting the function tries to use geolocation by default,
        if geolocation is not present it tries to use GCPs,
        if GCPs are not present it uses GeoTransform.

        If destination image has GCPs (provided in <dst_gcps>): fake GCPs for
        referencing line/pixel of SRC image and X/Y of DST image are created
        and added to the SRC image. After warping the dst_gcps are added to
        the warped VRT.

        Parameters
        -----------
        dst_srs : string
            WKT of the destination projection
        x_size, y_size : int
            dimentions of the destination rasetr
        geo_transform : tuple with 6 floats
            destination GDALGeoTransfrom
        resample_alg : int (GDALResampleAlg)
            0 : NearestNeighbour,
            1 : Bilinear,
            2 : Cubic,
            3 : CubicSpline,
            4 : Lancoz
        dst_gcps : list with GDAL GCPs
            GCPs of the destination image
        skip_gcps : int
            Using TPS can be very slow if the number of GCPs are large.
            If this parameter is given, only every [skip_gcp] GCP is used,
            improving calculation time at the cost of accuracy.
        block_size : int
            BlockSize to use for resampling. Larger blocksize reduces speed but increases accuracy.
        working_data_type : str
            'Float32', 'Int16', etc.
        resize_only : bool
            Create warped_vrt which will be used for resizing only?

        Returns
        --------
        warped_vrt : VRT object with warped dataset and with vrt


        """
        # VRT to be warped
        src_vrt = self.copy()

        # if destination GCPs are given: create and add fake GCPs to src
        dst_wkt = src_vrt._set_fake_gcps(dst_srs, dst_gcps, skip_gcps)

        if resize_only:
            src_vrt._set_geotransform_for_resize()
        else:
            src_vrt._set_gcps_geolocation_geotransform()

        # create Warped VRT GDAL Dataset
        warped_dataset = gdal.AutoCreateWarpedVRT(src_vrt.dataset, None, dst_wkt, resample_alg)

        # check if Warped VRT was created
        if warped_dataset is None:
            raise NansatGDALError('Cannot create warpedVRT!')

        # create VRT object from Warped VRT GDAL Dataset
        warped_vrt = VRT.copy_dataset(warped_dataset)

        # set x/y size, geo_transform, block_size
        warped_vrt._update_warped_vrt_xml(x_size, y_size, geo_transform, block_size, working_data_type)

        # apply thin-spline-transformation option
        if self.tps:
            warped_vrt.write_xml(warped_vrt.xml.replace('GCPTransformer', 'TPSTransformer'))

        # if given, add dst GCPs
        if len(dst_gcps) > 0:
            warped_vrt.dataset.SetGCPs(dst_gcps, dst_srs)
            warped_vrt._remove_geotransform()
            warped_vrt.dataset.SetProjection(str(''))

        # Copy self to warpedVRT
        warped_vrt.vrt = self.copy()

        # replace the reference from src_vrt to warped_vrt.vrt
        node0 = Node.create(str(warped_vrt.xml))
        node1 = node0.node('GDALWarpOptions')
        node1.node('SourceDataset').value = '/vsimem/' + str(os.path.basename(warped_vrt.vrt.filename))
        warped_vrt.write_xml(node0.rawxml())

        return warped_vrt

    def copyproj(self, filename):
        """ Copy geoloctation data from given VRT to a figure file

        Useful for adding geolocation information to figure
        files produced e.g. by Figure class, which contain no geolocation.
        Analogue to utility gdalcopyproj.py.

        Parameters
        -----------
        filename : string
            Name of file to which the geolocation data shall be written

        """
        figDataset = gdal.Open(filename, gdal.GA_Update)
        figDataset.SetGeoTransform(self.dataset.GetGeoTransform())
        figDataset.SetProjection(self.dataset.GetProjection())
        gcps = self.dataset.GetGCPs()
        if len(gcps) != 0:
            figDataset.SetGCPs(gcps, self.dataset.GetGCPProjection())
        figDataset = None  # Close and write output file

    def delete_band(self, band_num):
        """ Delete a band from the given VRT

        Parameters
        ----------
        band_num : int
            band number

        """
        node0 = Node.create(self.xml)
        node0.delNode('VRTRasterBand', options={'band': band_num})
        node0.delNode('BandMapping', options={'src': band_num})
        self.write_xml(node0.rawxml())

    def delete_bands(self, band_nums):
        """ Delete bands

        Parameters
        ----------
        bandNums : list
            elements are int

        """
        band_nums.sort(reverse=True)
        for i in band_nums:
            self.delete_band(i)

    def get_shifted_vrt(self, shift_degree):
        """ Roll data in bands westwards or eastwards

        Create shift_vrt which references self. Modify georeference
        of shift_vrt to account for the roll. Add as many bands as in self
        but for each band create two complex sources: for western
        and eastern parts. Keep self in shift_vrt.vrt

        Parameters
        ----------
        shift_degree : float
            rolling angle, how far east/west to roll

        Returns
        -------
        shift_vrt : VRT object with rolled bands

        """
        # Copy self into self.vrt
        shift_vrt = self.get_super_vrt()

        if shift_degree < 0:
            shift_degree += 360.0

        geo_transform = shift_vrt.vrt.dataset.GetGeoTransform()
        shift_pixel = int(shift_degree / float(geo_transform[1]))
        geo_transform = list(geo_transform)
        geo_transform[0] = round(geo_transform[0] + shift_degree, 3)
        new_east_border = geo_transform[0] + (geo_transform[1] * shift_vrt.dataset.RasterXSize)
        if new_east_border > 360.0:
            geo_transform[0] -= 360.0
        shift_vrt.dataset.SetGeoTransform(tuple(geo_transform))

        # read xml and create the node
        node0 = Node.create(shift_vrt.xml)

        # divide into two bands and switch the bands
        for i in range(len(node0.nodeList('VRTRasterBand'))):
            # create i-th 'VRTRasterBand' node
            node1 = node0.node('VRTRasterBand', i)
            # modify the 1st band
            shift_str = str(shift_pixel)
            size_str = str(shift_vrt.vrt.dataset.RasterXSize - shift_pixel)
            node1.node('ComplexSource').node('DstRect').replaceAttribute('xOff', shift_str)
            node1.node('ComplexSource').node('DstRect').replaceAttribute('xSize', size_str)
            node1.node('ComplexSource').node('SrcRect').replaceAttribute('xSize', size_str)

            # add the 2nd band
            clone_node = Node.create(node1.rawxml()).node('ComplexSource')
            clone_node.node('SrcRect').replaceAttribute('xOff', size_str)
            clone_node.node('DstRect').replaceAttribute('xOff', str(0))
            clone_node.node('SrcRect').replaceAttribute('xSize', shift_str)
            clone_node.node('DstRect').replaceAttribute('xSize', shift_str)

            # get VRTRasterBand with inserted ComplexSource
            node1 = node1.insert(clone_node.rawxml())
            node0.replaceNode('VRTRasterBand', i, node1)

        # write down XML contents
        shift_vrt.write_xml(node0.rawxml())

        return shift_vrt

    def get_sub_vrt(self, steps=1):
        """Returns sub-VRT from given depth

        Iteratively copy self.vrt into self until
        self.vrt is None or steps == 0

        Parameters
        -----------
        steps : int
            How many sub VRTs to restore

        Returns
        -------
        self : if no deeper VRTs found
        self.vrt : if deeper VRTs are found

        Notes
        --------
        self
        self.vrt

        """

        # check if self is the last valid (deepest) VRT
        if self.vrt is None:
            return self

        # check if required depth of restoration is met
        if steps == 0:
            return self

        # decrease the depth of restoration
        steps -= 1

        # return restored sub-VRT
        return self.vrt.get_sub_vrt(steps)

    def get_super_vrt(self):
        """Create vrt with subVRT

        copy of self in vrt.vrt and change references from vrt to vrt.vrt

        """
        # create new vrt that refers to a copy of self
        super_vrt = VRT.from_gdal_dataset(self.dataset, metadata=self.dataset.GetMetadata())
        super_vrt.vrt = self.copy()
        super_vrt.tps = self.tps

        # add bands to the new vrt
        for i in range(super_vrt.vrt.dataset.RasterCount):
            src = {'SourceFilename': super_vrt.vrt.filename, 'SourceBand': i + 1}
            dst = super_vrt.vrt.dataset.GetRasterBand(i + 1).GetMetadata()
            # remove PixelFunctionType from metadata to prevent its application
            if 'PixelFunctionType' in dst:
                dst.pop('PixelFunctionType')
            super_vrt.create_band(src, dst)
        super_vrt.dataset.FlushCache()

        return super_vrt

    def get_subsampled_vrt(self, new_raster_x_size, new_raster_y_size, resample_alg):
        """Create VRT and replace step in the source"""

        subsamp_vrt = self.get_super_vrt()

        # Get XML content from VRT-file
        node0 = Node.create(str(subsamp_vrt.xml))

        # replace rasterXSize in <VRTDataset>
        node0.replaceAttribute('rasterXSize', str(new_raster_x_size))
        node0.replaceAttribute('rasterYSize', str(new_raster_y_size))

        # replace xSize in <DstRect> of each source
        for iNode1 in node0.nodeList('VRTRasterBand'):
            for sourceName in ['ComplexSource', 'SimpleSource']:
                for iNode2 in iNode1.nodeList(sourceName):
                    iNodeDstRect = iNode2.node('DstRect')
                    iNodeDstRect.replaceAttribute('xSize', str(new_raster_x_size))
                    iNodeDstRect.replaceAttribute('ySize', str(new_raster_y_size))
            # if method=-1, overwrite 'ComplexSource' to 'AveragedSource'
            if resample_alg == -1:
                iNode1.replaceTag('ComplexSource', 'AveragedSource')
                iNode1.replaceTag('SimpleSource', 'AveragedSource')
                # if the values are complex number, give a warning
                if iNode1.getAttribute('dataType').startswith('C'):
                    warnings.warn(
                        'Band %s : The imaginary parts of complex numbers '
                        'are lost when resampling by averaging '
                        '(resample_alg=-1)' % iNode1.getAttribute('band'))

        # Write the modified elemements into VRT
        subsamp_vrt.write_xml(node0.rawxml())

        return subsamp_vrt

    def transform_points(self, col_vector, row_vector, dst2src=0,
                         dst_srs=NSR(), dst_ds=None, options=None):
        """Transform input pixel/line coordinates into lon/lat (or opposite)

        Parameters
        -----------
        col_vector, row_vector : lists
            X and Y coordinates with any coordinate system
        dst2src : 0 or 1
            1 for inverse transformation, 0 for forward transformation.
        dst_srs : NSR
            destination SRS.
        dst_ds : GDAL Dataset
            destination dataset. The default is None.
            It means transform ownPixLin <--> ownXY.
        option : string
            if 'METHOD=GEOLOC_ARRAY', specify here.

        Returns
        --------
        lon_vector, lat_vector : numpy arrays
            X and Y coordinates in degree of lat/lon

        """
        # get source SRS (either Projection or GCPProjection)
        src_wkt = self.get_projection()

        # prepare options
        if options is None:
            options = ['SRC_SRS=' + src_wkt, 'DST_SRS=' + dst_srs.wkt]
            # add TPS method if we have GCPs and self.tps is True
            if self.tps and len(self.dataset.GetGCPs()) > 0:
                options.append('METHOD=GCP_TPS')

        # create transformer
        transformer = gdal.Transformer(self.dataset, dst_ds, options)

        # convert lists with X,Y coordinates to 2D numpy array
        xy = np.array([col_vector, row_vector]).transpose()

        # transfrom coordinates
        lonlat = transformer.TransformPoints(dst2src, xy)[0]

        # convert return to lon,lat vectors
        lonlat = np.array(lonlat)
        if lonlat.shape[0] > 0:
            lon_vector = lonlat[:, 0]
            lat_vector = lonlat[:, 1]
        else:
            lon_vector, lat_vector = [], []

        return lon_vector, lat_vector

    def get_projection(self):
        """Get projection from the dataset.

        Uses gdal.Dataset.GetProjection() or gdal.Dataset.GetGCPProjection(). If both return an
        empty string, a NansatProjectionError is raised.

        Returns
        -------
        projection : projection or GCPprojection

        Raises
        -------
        NansatProjectionError : occurs when the projection is empty.

        """
        # get projection or GCPProjection
        projection = self.dataset.GetProjection()
        if projection == '':
            projection = self.dataset.GetGCPProjection()

        if not projection:
            raise NansatProjectionError

        return projection

    def get_resized_vrt(self, x_size, y_size, resample_alg=1):

        """ Resize VRT

        Create Warped VRT with modified RasterXSize, RasterYSize, GeoTransform.
        The returned VRT object has a copy of its original source VRT in its
        own vrt object (e.g. warpedVRT.vrt = originalVRT.copy()).

        Parameters
        -----------
        x_size, y_size : int
            new size of the VRT object
        resample_alg : GDALResampleAlg
            see also gdal.AutoCreateWarpedVRT

        Returns
        --------
        warped_vrt : Resized VRT object

        """
        # modify GeoTransform: set resolution from new X/Y size
        geo_transform = (0.5,
                        float(self.dataset.RasterXSize - 1.0) / float(x_size),
                        0,
                        self.dataset.RasterYSize - 0.5,
                        0,
                        - float(self.dataset.RasterYSize - 1.0) / float(y_size))

        # update size and GeoTranform in XML of the warped VRT object
        warped_vrt = self.get_warped_vrt(None, x_size, y_size, geo_transform, resample_alg,
                                         resize_only=True)

        return warped_vrt

    def reproject_GCPs(self, dst_srs):
        """Reproject all GCPs to a new spatial reference system

        Necessary before warping an image if the given GCPs
        are in a coordinate system which has a singularity
        in (or near) the destination area (e.g. poles for lonlat GCPs)

        Parameters
        ----------
        dst_srs : proj4, WKT, NSR, EPSG
            Destiination SRS given as any NSR input parameter

        Notes
        --------
        Reprojects all GCPs to new SRS and updates GCPProjection
        """
        # transform coordinates of original GCPs
        dst_srs = NSR(dst_srs)
        src_srs = NSR(self.dataset.GetGCPProjection())
        # make three tuples with X, Y and Z values of GCPs
        src_points = list(zip(*[(gcp.GCPX, gcp.GCPY, gcp.GCPZ) for gcp in self.dataset.GetGCPs()]))
        dst_points = VRT.transform_coordinates(src_srs, src_points, dst_srs)
        # create new GCPs
        dst_gcps = []
        for p in zip(self.dataset.GetGCPs(), *dst_points):
            dst_gcp = gdal.GCP(p[1], p[2], p[3],
                               p[0].GCPPixel,  p[0].GCPLine, p[0].Info, p[0].Id)
            dst_gcps.append(dst_gcp)
        # Update dataset
        self.dataset.SetGCPs(dst_gcps, dst_srs.wkt)
        self.dataset.FlushCache()

    @staticmethod
    def transform_coordinates(src_srs, src_points, dst_srs):
        """ Transform coordinates of points from one spatial reference to another

        Parameters
        ----------
        src_srs : nansat.NSR
            Source spatial reference system
        src_points : tuple of two or three N-D arrays
            Coordinates of points in the source spatial reference system. A tuple with (X, Y) or
            (X, Y, Z) coordinates arrays. Each coordinate Each array can be a list, 1D, 2D, N-D
            array.
        dst_srs : nansat.NSR
            Destination spatial reference

        Returns
        -------
        dst_points : tuple of two or three N-D arrays
            Coordinates of points in the destination spatial reference system. A tuple with (X, Y) or
            (X, Y, Z) coordinates arrays. Each coordinate Each array can be 1D, 2D, N-D
            array. Shape of output arrays corrrrespond to shape of inputs.

        """
        transformer = osr.CoordinateTransformation(src_srs, dst_srs)
        src_shape = np.array(src_points[0]).shape
        src_points = list(zip(*[np.array(xyz).flatten() for xyz in src_points]))
        dst_points = transformer.TransformPoints(src_points)
        dst_x, dst_y, dst_z = np.array(list(zip(*dst_points)))
        return dst_x.reshape(src_shape), dst_y.reshape(src_shape), dst_z.reshape(src_shape)

    def set_offset_size(self, axis, offset, size):
        """Set offset and  size in VRT dataset and band attributes"""
        node0 = Node.create(str(self.xml))

        # change size
        node0.node('VRTDataset').replaceAttribute('raster%sSize'%str(axis).upper(), str(size))

        # replace x/y-Off and x/y-Size
        #   in <SrcRect> and <DstRect> of each source
        for iNode1 in node0.nodeList('VRTRasterBand'):
            iNode2 = iNode1.node('ComplexSource')

            iNode3 = iNode2.node('SrcRect')
            iNode3.replaceAttribute('%sOff'%str(axis).lower(), str(offset))
            iNode3.replaceAttribute('%sSize'%str(axis).lower(), str(size))

            iNode3 = iNode2.node('DstRect')
            iNode3.replaceAttribute('%sSize'%str(axis).lower(), str(size))

        # write modified XML
        self.write_xml(node0.rawxml())

    def shift_cropped_gcps(self, x_offset, x_size, y_offset, y_size):
        """Modify GCPs to fit the size/offset of cropped image"""
        gcps = self.dataset.GetGCPs()
        if len(gcps) == 0:
            return

        dst_gcps = []
        i = 0
        # keep current GCPs
        for igcp in gcps:
            if (0 < igcp.GCPPixel - x_offset < x_size and 0 < igcp.GCPLine - y_offset < y_size):
                i += 1
                dst_gcps.append(gdal.GCP(igcp.GCPX, igcp.GCPY, 0,
                                         igcp.GCPPixel - x_offset,
                                         igcp.GCPLine - y_offset, str(''), str(i)))
        n_gcps = i
        if n_gcps < 100:
            # create new 100 GPCs (10 x 10 regular matrix)
            pix_array, lin_array = np.mgrid[0:x_size:10j, 0:y_size:10j]
            pix_array = pix_array.flatten() + x_offset
            lin_array = lin_array.flatten() + y_offset

            lon_array, lat_array = self.vrt.transform_points(pix_array, lin_array,
                                                dst_srs=NSR(self.vrt.dataset.GetGCPProjection()))

            for i in range(len(lon_array)):
                dst_gcps.append(gdal.GCP(lon_array[i], lat_array[i], 0,
                                         pix_array[i] - x_offset,
                                         lin_array[i] - y_offset,
                                         str(''), str(n_gcps+i+1)))

        # set new GCPss
        self.dataset.SetGCPs(dst_gcps, self.vrt.dataset.GetGCPProjection())
        # remove geotranform which was automatically added
        self._remove_geotransform()

    def shift_cropped_geo_transform(self, x_offset, x_size, y_offset, y_size):
        """Modify GeoTransform to fit the size/offset of the cropped image"""
        if len(self.vrt.dataset.GetGCPs()) != 0:
            return
        # shift upper left corner coordinates
        geoTransfrom = self.dataset.GetGeoTransform()
        geoTransfrom = list(map(float, geoTransfrom))
        geoTransfrom[0] += geoTransfrom[1] * x_offset
        geoTransfrom[3] += geoTransfrom[5] * y_offset
        self.dataset.SetGeoTransform(geoTransfrom)
        self.dataset.FlushCache()

    @staticmethod
    def read_vsi(filename):
        """Read text from input <filename:str> using VSI and return <content:str>."""
        # open
        vsi_file = gdal.VSIFOpenL(str(filename), str('r'))
        # get file size
        gdal.VSIFSeekL(vsi_file, 0, 2)
        vsi_file_size = gdal.VSIFTellL(vsi_file)
        # fseek to start again
        gdal.VSIFSeekL(vsi_file, 0, 0)
        # read
        vsi_file_content = gdal.VSIFReadL(vsi_file_size, 1, vsi_file)
        gdal.VSIFCloseL(vsi_file)
        return str(vsi_file_content.decode())

    @staticmethod
    def _make_source_bands_xml(src_in):
        """Check parameters of band source, set defaults and generate XML for VRT

        Parameters
        -------
            src_in : dict
                dict with band source parameters (SourceFilename, SourceBand, etc)
        Returns
        -------
            src : dict
                updated dict with XML entry
        """
        # check if SourceFilename is given
        if 'SourceFilename' not in src_in:
            raise KeyError('SourceFilename must be available in dict:src_in')
        # set default values
        src = {'SourceBand': 1,
               'LUT': '',
               'NODATA': '',
               'SourceType': 'ComplexSource',
               'ScaleRatio': 1.0,
               'ScaleOffset': 0.0}
        src.update(src_in)

        # find DataType of source (if not given in src)
        if src['SourceBand'] > 0 and 'DataType' not in src:
            raster_band = gdal.Open(src['SourceFilename']).GetRasterBand(src['SourceBand'])
            src['DataType'] = raster_band.DataType

        if 'xSize' not in src or 'ySize' not in src:
            ds = gdal.Open(src['SourceFilename'])
            src['xSize'] = ds.RasterXSize
            src['ySize'] = ds.RasterYSize

        # create XML for each source
        src['XML'] = VRT.COMPLEX_SOURCE_XML.substitute(
            Dataset=src['SourceFilename'],
            SourceBand=src['SourceBand'],
            SourceType=src['SourceType'],
            NODATA=src['NODATA'],
            ScaleOffset=src['ScaleOffset'],
            ScaleRatio=src['ScaleRatio'],
            LUT=src['LUT'],
            xSize=src['xSize'],
            ySize=src['ySize'],
            xOff=src.get('xOff', 0),
            yOff=src.get('yOff', 0),)

        return src

    @staticmethod
    def _set_add_band_options(srcs, dst):
        """Generate options for gdal.AddBand based on input band src and dst parameters"""
        if 'PixelFunctionType' in dst and len(dst['PixelFunctionType']) > 0:
            # in case of PixelFunction
            options = ['subClass=VRTDerivedRasterBand',
                       'PixelFunctionType=%s' % dst['PixelFunctionType']]
            if 'SourceTransferType' in dst:
                options.append('SourceTransferType=%s' % dst['SourceTransferType'])
        elif len(srcs) == 1 and srcs[0]['SourceBand'] == 0:
            # in case of VRTRawRasterBand
            options = ['subclass=VRTRawRasterBand',
                       'SourceFilename=%s' % srcs[0]['SourceFilename'],
                       'ImageOffset=%d' % srcs[0]['ImageOffset'],
                       'PixelOffset=%d' % srcs[0]['PixelOffset'],
                       'LineOffset=%d' % srcs[0]['LineOffset'],
                       'ByteOrder=%s' % srcs[0]['ByteOrder']]
        else:
            # in common case
            options = []

        return options

    @staticmethod
    def _lonlat2gcps(lon, lat, n_gcps=100, **kwargs):
        """ Create list of GCPs from given grids of latitude and longitude

        take <n_gcps> regular pixels from inpt <lat> and <lon> grids
        Create GCPs from these pixels
        Create latlong GCPs projection

        Parameters
        -----------
        lat : Numpy grid
            array of latitudes
        lon : Numpy grid
            array of longitudes (should be the same size as lat)
        n_gcps : int, optional, default = 100
            number of GCPs to create

        Returns
        --------
        gcsp : List with GDAL GCPs

        """
        # estimate step of GCPs
        gcp_size = np.sqrt(n_gcps)
        step0 = max(1, int(float(lat.shape[0]) / gcp_size))
        step1 = max(1, int(float(lat.shape[1]) / gcp_size))

        # generate list of GCPs
        gcps = []
        k = 0
        for i0 in range(0, lat.shape[0], step0):
            for i1 in range(0, lat.shape[1], step1):
                # create GCP with X,Y,pixel,line from lat/lon matrices
                gcp = gdal.GCP(float(lon[i0, i1]), float(lat[i0, i1]), 0, i1, i0)
                gcps.append(gcp)
                k += 1

        return gcps

    @staticmethod
    def _put_metadata(raster_band, metadata_dict):
        """ Put all metadata into a raster band

        Take metadata from metadataDict and put to the GDAL Raster Band

        Parameters
        ----------
        rasterBand : GDALRasterBand
            destination band without metadata

        metadataDict : dictionary
            keys are names of metadata, values are values

        Returns
        --------
        rasterBand : GDALRasterBand
            destination band with metadata

        """
        for key in metadata_dict:
            try:
                meta_value = str(metadata_dict[key])
                meta_key = str(key)
            except UnicodeEncodeError:
                self.logger.error('Cannot add %s to metadata' % key)
            else:
                raster_band.SetMetadataItem(meta_key, meta_value)

        return raster_band

    @staticmethod
    def _make_filename(extention='vrt', nomem=False):
        """Create random VSI file name

        Parameters
        ----------
        extention : string
            extension of the file

        Returns
        -------
        random file name

        """
        if nomem:
            fd, filename = tempfile.mkstemp(suffix='.'+extention)
            os.close(fd)
        else:
            all_chars = ascii_uppercase + digits
            random_chars = ''.join(choice(all_chars) for x in range(10))
            filename = '/vsimem/%s.%s' % (random_chars, extention)
        return filename

    @staticmethod
    def _get_dst_band_data_type(srcs, dst):
        """Get destination dataType based on source and destnation band dictionaries"""
        if 'dataType' in dst:                                               # data type is preset
            data_type = dst['dataType']
        elif (len(srcs) > 1 or                                                     # pixel function
              ('ScaleRatio' in srcs[0] and float(srcs[0]['ScaleRatio']) != 1.0) or # scaling applied
              ('LUT' in srcs[0] and len(srcs[0]['LUT'])) > 0 or                    # if LUT exists
              'DataType' not in srcs[0]):               # if source band not available
            data_type = gdal.GDT_Float32
        else:
            # otherwise take the DataType from source
            data_type = srcs[0]['DataType']
        return data_type

    @staticmethod
    def _remove_strings_in_metadata_keys(gdal_metadata, rm_strings):
        """Remove unwanted metadata"""

        new_meta = {}
        for key, value in gdal_metadata.items():
            for rms in rm_strings:
                if rms in key:
                    key = key.replace(rms, '')
            new_meta[key] = value

        return new_meta



# Name:         mapper_globcurrent_online.py
# Purpose:      Nansat mapping for GLOBCURRENT data, stored online in HYRAX
# Author:       Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html
import os

from dateutil.parser import parse
import json
import numpy as np

import pythesint as pti

from nansat.nsr import NSR
from nansat.mappers.opendap import Opendap

class Mapper(Opendap):
    ''' VRT with mapping of WKV for NCEP GFS '''
    #http://www.ifremer.fr/opendap/cerdap1/globcurrent/v2.0/global_025_deg/total_hs/2010/001/20100101000000-GLOBCURRENT-L4-CUReul_hs-ALT_SUM-v02.0-fv01.0.nc
    #http://tds0.ifremer.fr/thredds/dodsC/CLS-L4-CUREUL_HS-ALT_SUM-V02.0_FULL_TIME_SERIE
    baseURLs = ['http://www.ifremer.fr/opendap/cerdap1/globcurrent/v2.0/']
    timeVarName = 'time'
    xName = 'lon'
    yName = 'lat'
    timeCalendarStart = '1950-01-01'

    srcDSProjection = NSR().wkt

    def __init__(self, filename, gdalDataset, gdalMetadata,
                 date=None, ds=None, bands=None, cachedir=None,
                 **kwargs):
        ''' Create NCEP VRT
        Parameters:
            filename : URL
            date : str
                2010-05-01
            ds : netCDF.Dataset
                previously opened dataset

        '''
        self.test_mapper(filename)
        fname = os.path.split(filename)[1]
        date = '%s-%s-%sT%s:00Z' % (fname[0:4], fname[4:6], fname[6:8], fname[8:10])

        self.create_vrt(filename, gdalDataset, gdalMetadata, date, ds, bands, cachedir)

        # add instrument and platform
        #instr = pti.get_gcmd_instrument('active remote sensing')
        #pltfr = pti.get_gcmd_platform('Earth Observation Satellites')
        pltfr = pti.get_gcmd_platform('JASON-1')
        instr = pti.get_gcmd_instrument('JASON-2 RADAR ALTIMETER')

        self.dataset.SetMetadataItem('instrument',  json.dumps(instr))
        self.dataset.SetMetadataItem('platform',    json.dumps(pltfr))
        self.dataset.SetMetadataItem('Data Center', 'FR/IFREMER/CERSAT')
        self.dataset.SetMetadataItem('Entry Title', 'GLOBCURRENT')

    def convert_dstime_datetimes(self, dsTime):
        ''' Convert time variable to np.datetime64 '''
        dsDatetimes = np.array([(np.datetime64(self.timeCalendarStart).astype('M8[s]') +
                                 np.timedelta64(int(day), 'D').astype('m8[s]') +
                                 np.timedelta64(int(24*(day - int(day))), 'h').astype('m8[s]'))
                                for day in dsTime]).astype('M8[s]')

        return dsDatetimes
